#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import pytest
from ezdxf.addons import binpacking as bp


def test_single_bin_single_item():

    packer = bp.Packer()
    box = packer.add_bin("B0", 1, 1, 1)
    packer.add_item("I0", 1, 1, 1, 1)
    packer.pack()
    assert len(box.items) == 1
    assert len(packer.unfitted_items) == 0


@pytest.mark.parametrize(
    "w,h,d",
    [
        (3, 1, 1),
        (1, 3, 1),
        (1, 1, 3),
    ],
)
def test_single_bin_multiple_items(w, h, d):
    packer = bp.Packer()
    box = packer.add_bin("B0", w, h, d)
    for index in range(max(w, h, d)):
        packer.add_item(f"I{index}", 1, 1, 1, 1)
    packer.pack()
    assert len(box.items) == 3
    assert len(packer.unfitted_items) == 0


def test_single_bin_different_sized_items():
    packer = bp.Packer()
    box = packer.add_bin("B0", 3, 3, 1)
    packer.add_item("I0", 1, 1, 1, 1)
    packer.add_item("I1", 2, 1, 1, 1)
    packer.add_item("I2", 3, 1, 1, 1)
    packer.pack()
    assert len(box.items) == 3
    assert len(packer.unfitted_items) == 0


def test_empty_packer():
    packer = bp.Packer()
    assert packer.get_capacity() == 0.0
    assert packer.get_total_volume() == 0.0
    assert packer.get_total_weight() == 0.0
    assert packer.get_fill_ratio() == 0.0
    assert len(packer.unfitted_items) == 0


def test_empty_box():
    box = bp.Box("box", 1, 1, 1)
    assert box.get_capacity() == 1.0
    assert box.get_fill_ratio() == 0.0


def test_cannot_create_zero_sized_box():
    with pytest.raises(ValueError):
        bp.Box("box", 0, 2, 3)
    with pytest.raises(ValueError):
        bp.Box("box", 1, 0, 3)
    with pytest.raises(ValueError):
        bp.Box("box", 1, 2, 0)


def test_forced_zero_sized_box():
    box = bp.Box("box", 1, 2, 3)
    box.width = 0
    assert box.get_capacity() == 0.0
    assert box.get_fill_ratio() == 0.0


SMALL_ENVELOPE = ("small-envelope", 11.5, 6.125, 0.25, 10)
LARGE_ENVELOPE = ("large-envelope", 15.0, 12.0, 0.75, 15)
SMALL_BOX = ("small-box", 8.625, 5.375, 1.625, 70.0)
MEDIUM_BOX = ("medium-box", 11.0, 8.5, 5.5, 70.0)
MEDIUM_BOX2 = ("medium-box-2", 13.625, 11.875, 3.375, 70.0)
LARGE_BOX = ("large-box", 12.0, 12.0, 5.5, 70.0)
LARGE_BOX2 = ("large-box-2", 23.6875, 11.75, 3.0, 70.0)

ALL_BINS = [
    SMALL_ENVELOPE,
    LARGE_ENVELOPE,
    SMALL_BOX,
    MEDIUM_BOX,
    MEDIUM_BOX2,
    LARGE_BOX,
    LARGE_BOX2,
]


@pytest.fixture
def packer():
    packer = bp.Packer()
    packer.add_item("50g [powder 1]", 3.9370, 1.9685, 1.9685, 1)
    packer.add_item("50g [powder 2]", 3.9370, 1.9685, 1.9685, 2)
    packer.add_item("50g [powder 3]", 3.9370, 1.9685, 1.9685, 3)
    packer.add_item("250g [powder 4]", 7.8740, 3.9370, 1.9685, 4)
    packer.add_item("250g [powder 5]", 7.8740, 3.9370, 1.9685, 5)
    packer.add_item("250g [powder 6]", 7.8740, 3.9370, 1.9685, 6)
    packer.add_item("250g [powder 7]", 7.8740, 3.9370, 1.9685, 7)
    packer.add_item("250g [powder 8]", 7.8740, 3.9370, 1.9685, 8)
    packer.add_item("250g [powder 9]", 7.8740, 3.9370, 1.9685, 9)
    return packer


def pack(packer, box, pick):
    packer.add_bin(*box)
    packer.pack(pick)
    return packer.bins[0]


class TestExampleSmallerFirst:
    @staticmethod
    def pack(packer, box):
        return pack(packer, box, bp.PickStrategy.SMALLER_FIRST)

    @pytest.mark.parametrize("box", [SMALL_ENVELOPE, LARGE_ENVELOPE, SMALL_BOX])
    def test_small_bins(self, packer, box):
        b0 = self.pack(packer, box)
        assert len(b0.items) == 0
        assert b0.get_total_weight() == 0
        assert b0.get_total_volume() == 0.0
        assert b0.get_fill_ratio() == 0.0

    def test_medium_box(self, packer):
        b0 = self.pack(packer, MEDIUM_BOX)
        assert len(b0.items) == 6
        assert b0.get_total_weight() == 21
        assert b0.get_total_volume() == pytest.approx(228.83766732374997)
        assert b0.get_fill_ratio() == pytest.approx(0.44499303320126393)

    def test_medium_box2(self, packer):
        b0 = self.pack(packer, MEDIUM_BOX2)
        assert len(b0.items) == 6
        assert b0.get_total_weight() == 21
        assert b0.get_total_volume() == pytest.approx(228.83766732374997)
        assert b0.get_fill_ratio() == pytest.approx(0.4190671376138204)

    def test_large_box(self, packer):
        b0 = self.pack(packer, LARGE_BOX)
        assert len(b0.items) == 9
        assert b0.get_total_weight() == 45
        assert b0.get_total_volume() == pytest.approx(411.90780118274995)
        assert b0.get_fill_ratio() == pytest.approx(0.5200856075539773)

    def test_large_box2(self, packer):
        b0 = self.pack(packer, LARGE_BOX2)
        assert len(b0.items) == 9
        assert b0.get_total_weight() == 45
        assert b0.get_total_volume() == pytest.approx(411.90780118274995)
        assert b0.get_fill_ratio() == pytest.approx(0.4933119870449671)


class TestExampleBiggerFirst:
    @staticmethod
    def pack(packer, box):
        return pack(packer, box, bp.PickStrategy.BIGGER_FIRST)

    @pytest.mark.parametrize("box", [SMALL_ENVELOPE, LARGE_ENVELOPE, SMALL_BOX])
    def test_small_bins(self, packer, box):
        b0 = self.pack(packer, box)
        assert len(b0.items) == 0
        assert b0.get_total_weight() == 0
        assert b0.get_total_volume() == 0.0
        assert b0.get_fill_ratio() == 0.0

    def test_medium_box(self, packer):
        b0 = self.pack(packer, MEDIUM_BOX)
        assert len(b0.items) == 5
        assert b0.get_total_weight() == 30
        assert b0.get_total_volume() == pytest.approx(305.116889765)
        assert b0.get_fill_ratio() == pytest.approx(0.593324044268352)

    def test_medium_box2(self, packer):
        b0 = self.pack(packer, MEDIUM_BOX2)
        assert len(b0.items) == 6
        assert b0.get_total_weight() == 25
        assert b0.get_total_volume() == pytest.approx(274.6052007884999)
        assert b0.get_fill_ratio() == pytest.approx(0.5028805651365844)

    def test_large_box(self, packer):
        b0 = self.pack(packer, LARGE_BOX)
        assert len(b0.items) == 9
        assert b0.get_total_weight() == 45
        assert b0.get_total_volume() == pytest.approx(411.90780118274995)
        assert b0.get_fill_ratio() == pytest.approx(0.5200856075539773)

    def test_large_box2(self, packer):
        b0 = self.pack(packer, LARGE_BOX2)
        assert len(b0.items) == 9
        assert b0.get_total_weight() == 45
        assert b0.get_total_volume() == pytest.approx(411.90780118274995)
        assert b0.get_fill_ratio() == pytest.approx(0.4933119870449671)


@pytest.mark.parametrize(
    "item",
    [
        bp.Item([1, 2, 3], 1, 2, 3, 4),
        bp.FlatItem([1, 2, 3], 1, 2, 4),
    ],
)
def test_copy_item(item):
    assert item.bbox is not None  # trigger bounding box update
    item2 = item.copy()
    assert item.payload is item2.payload, "should reference the same object"
    assert item.get_dimension() == item2.get_dimension()
    assert item.position == item2.position
    assert item.weight == item2.weight
    assert item.bbox is item2.bbox


class TestItemTransformation:
    """Transformation of the source entity located with the minimum extension
    corner of its bounding box in (0, 0, 0) in width (x), height (y) and depth (z)
    orientation to the final location including the required rotation.
    """

    @pytest.fixture
    def item(self):
        i = bp.Item("box", 2, 3, 4)
        i.position = (3, 2, 1)  # target location
        return i

    def test_whd_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.WHD  # width, height, depth
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((3, 2, 1))

    def test_hwd_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.HWD  # height, width, depth
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((6, 2, 1))

    def test_hdw_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.HDW  # height, depth, width
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((6, 6, 1))

    def test_dhw_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.DHW  # depth, height, width
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((7, 2, 1))

    def test_dwh_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.DWH  # depth, width, height
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((3, 2, 1))

    def test_wdh_rotation(self, item: bp.Item):
        item.rotation_type = bp.RotationType.WDH  # width, depth, height
        m = item.get_transformation()
        assert m.transform((0, 0, 0)).isclose((3, 6, 1))


def test_copy_box():
    box = bp.Box("NAME", 1, 2, 3, 4)
    box.items = [1, 2, 3]
    box.unfitted_items = [4, 5, 6]
    box2 = box.copy()

    assert box.name == box2.name
    assert box.width == box2.width
    assert box.height == box2.height
    assert box.max_weight == box2.max_weight
    assert box.items is not box2.items, "expected shallow copy"


def test_copy_packer(packer):
    packer2 = packer.copy()
    assert packer.bins is not packer2.bins, "expected shallow copy"
    assert len(packer.bins) == len(packer2.bins)
    assert packer.items is not packer2.items, "expected shallow copy"
    assert len(packer.items) == len(packer2.items)


def test_cannot_copy_packed_packer(packer):
    packer.pack(pick=bp.PickStrategy.BIGGER_FIRST)
    with pytest.raises(TypeError):
        packer.copy()


def test_cannot_copy_packer_with_non_empty_bins(packer):
    box = bp.Bin(*LARGE_BOX)
    packer.append_bin(box)
    box.items.append(0)  # type: ignore
    with pytest.raises(TypeError):
        packer.copy()


def test_cannot_append_bins_to_packed_packer(packer):
    packer.pack(pick=bp.PickStrategy.BIGGER_FIRST)
    with pytest.raises(TypeError):
        packer.append_bin(bp.Bin(*LARGE_BOX))


def test_cannot_append_non_empty_bins(packer):
    box = bp.Bin(*LARGE_BOX)
    box.items.append(0)  # type: ignore
    with pytest.raises(TypeError):
        packer.append_bin(box)


def test_cannot_append_items_to_packed_packer(packer):
    packer.pack(pick=bp.PickStrategy.BIGGER_FIRST)
    with pytest.raises(TypeError):
        packer.append_item(bp.Item(0, 0, 0, 0))


def test_random_shuffle_interface(packer):
    packer.add_bin(*LARGE_BOX2)
    best = bp.shuffle_pack(packer, 2)
    assert best.get_fill_ratio() > 0.0


def test_random_shuffle_raise_exception_for_invalid_attempts(packer):
    with pytest.raises(ValueError):
        bp.shuffle_pack(packer, 0)


def test_pack_item_subset(packer):
    packer.add_bin(*LARGE_BOX)
    bp.pack_item_subset(packer, picker=[0, 1, 1, 1])
    assert len(packer.bins[0]) == 3
    assert packer.get_total_weight() == 9
    assert len(packer.unfitted_items) == 6


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
# mypy: ignore_errors=True
from ezdxf._options import DRAWING_ADDON, options

# Qt compatibility layer: all Qt imports from ezdxf.addons.xqt
PYSIDE6 = False
TRY_PYSIDE6 = options.get_bool(DRAWING_ADDON, "try_pyside6", True)
PYQT5 = False
TRY_PYQT5 = options.get_bool(DRAWING_ADDON, "try_pyqt5", True)

if TRY_PYSIDE6:
    try:
        from PySide6 import QtGui, QtCore, QtWidgets
        from PySide6.QtWidgets import (
            QFileDialog,
            QInputDialog,
            QMessageBox,
            QTableView,
            QTreeView,
            QListView,
        )
        from PySide6.QtCore import (
            QAbstractTableModel,
            QStringListModel,
            QFileSystemWatcher,
            QModelIndex,
            QPointF,
            QSettings,
            QSize,
            Qt,
            Signal,
            Slot,
        )
        from PySide6.QtGui import (
            QAction,
            QColor,
            QPainterPath,
            QStandardItem,
            QStandardItemModel,
        )

        PYSIDE6 = True
        print("using Qt binding: PySide6")
    except ImportError:
        pass

# PyQt5 is just a fallback
if TRY_PYQT5 and not PYSIDE6:
    try:
        from PyQt5 import QtGui, QtCore, QtWidgets
        from PyQt5.QtCore import pyqtSignal as Signal
        from PyQt5.QtCore import pyqtSlot as Slot
        from PyQt5.QtWidgets import (
            QAction,
            QFileDialog,
            QInputDialog,
            QMessageBox,
            QTableView,
            QTreeView,
            QListView,
        )
        from PyQt5.QtCore import (
            QAbstractTableModel,
            QStringListModel,
            QFileSystemWatcher,
            QModelIndex,
            QPointF,
            QSettings,
            QSize,
            Qt,
        )
        from PyQt5.QtGui import (
            QColor,
            QPainterPath,
            QStandardItem,
            QStandardItemModel,
        )

        PYQT5 = True
        print("using Qt binding: PyQt5")
    except ImportError:
        pass

if not (PYSIDE6 or PYQT5):
    raise ImportError("no Qt binding found, tried PySide6 and PyQt5")



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.tools import pattern


def test_load_iso_pattern():
    p = pattern.load()
    assert p["ANSI31"][0] == [
        45.0,
        (0.0, 0.0),
        (-2.2450640303, 2.2450640303),
        [],
    ]


def test_load_scaled_iso_pattern():
    p = pattern.load(factor=2)
    assert p["ANSI31"][0] == [
        45.0,
        (0.0, 0.0),
        (-4.4901280606, 4.4901280606),
        [],
    ]


def test_load_imperial_pattern():
    p = pattern.load(measurement=0)
    assert p["ANSI31"][0] == [
        45.0,
        (0.0, 0.0),
        (-0.0883883476, 0.0883883476),
        [],
    ]


def test_scale_pattern():
    p = pattern.load()
    ansi31 = p["ANSI31"]
    s = pattern.scale_pattern(ansi31, 2, angle=90)

    angle, base, offset, lines = s[0]
    assert angle == 135
    assert base == (0, 0)
    assert offset == (-4.4901280606, -4.4901280606)


def test_scale_all_pattern():
    r = pattern.scale_all(pattern.ISO_PATTERN)
    assert len(r) == len(pattern.ISO_PATTERN)


TEST_PATTERN = """; Hatch Patterns adapted to ISO scaling

;; Note: Dummy pattern description used for 'Solid fill'.
*SOLID, Solid fill
45, 0,0, 0,.125
*ANSI31, ANSI Iron, Brick, Stone masonry
45, 0,0, 0,3.175 
*ANSI32, ANSI Steel
45, 0,0, 0,9.525 
45, 4.490128053,0, 0,9.525 
*ANSI33, ANSI Bronze, Brass, Copper
45, 0,0, 0,6.35 
45, 4.490128053,0, 0,6.35, 3.175,-1.5875
*ANSI34, ANSI Plastic, Rubber
45, 0,0, 0,19.05 
45, 4.490128053,0, 0,19.05 
45, 8.9802561314,0, 0,19.05 
45, 13.4703841844,0, 0,19.05 
*ANSI35, ANSI Fire brick, Refractory material
45, 0,0, 0,6.35 
45, 4.490128053,0, 0,6.35, 7.9375,-1.5875,0,-1.5875
*ANSI36, ANSI Marble, Slate, Glass
45, 0,0, 5.55625,3.175, 7.9375,-1.5875,0,-1.5875
*ANSI37, ANSI Lead, Zinc, Magnesium, Sound/Heat/Elec Insulation
45, 0,0, 0,3.175 
135, 0,0, 0,3.175 
*ANSI38, ANSI Aluminum
45, 0,0, 0,3.175 
135, 0,0, 6.35,3.175, 7.9375,-4.7625

"""


def test_parse_pattern_file():
    result = pattern.parse(TEST_PATTERN)
    assert len(result) == 9
    assert result["SOLID"] == [
        [45.0, (0.0, 0.0), (-0.0883883476, 0.0883883476), []]
    ]
    assert result["ANSI33"] == [
        [45.0, (0.0, 0.0), (-4.4901280605, 4.4901280605), []],
        [
            45.0,
            (4.490128053, 0.0),
            (-4.4901280605, 4.4901280605),
            [3.175, -1.5875],
        ],
    ]


def analyse(name):
    return pattern.PatternAnalyser(pattern.ISO_PATTERN[name])


class TestPatternAnalyser:
    @pytest.mark.parametrize("name", ["ISO02W100", "DASH", "CLAY", "FLEXIBLE"])
    def test_has_horizontal_lines(self, name):
        result = analyse(name)
        assert result.has_angle(0) is True

    @pytest.mark.parametrize(
        "name",
        [
            "GOST_WOOD",
            "V_ZINC",
        ],
    )
    def test_has_vertical_lines(self, name):
        result = analyse(name)
        assert result.has_angle(90) is True

    @pytest.mark.parametrize(
        "name", ["ANSI31", "BRICK_INSULATING", "BUTTERFLY", "CROSSES"]
    )
    def test_has_45_deg_lines(self, name):
        result = analyse(name)
        assert result.has_angle(45) is True

    @pytest.mark.parametrize("name", ["BUTTERFLY", "CROSSES"])
    def test_has_135_deg_lines(self, name):
        result = analyse(name)
        assert result.has_angle(135) is True

    @pytest.mark.parametrize("name", ["DIAMONDS", "ESCHER"])
    def test_has_60_deg_lines(self, name):
        result = analyse(name)
        assert result.has_angle(60) is True

    @pytest.mark.parametrize("name", ["ANSI31", "BRICK_EXISTING"])
    def test_all_45_deg_lines(self, name):
        result = analyse(name)
        assert result.all_angles(45) is True

    @pytest.mark.parametrize("name", ["ANSI31", "BRICK_EXISTING"])
    def test_all_solid_lines(self, name):
        result = analyse(name)
        assert result.all_solid_lines() is True

    def test_analyse_ansi31(self):
        result = analyse("ANSI31")
        assert result.has_angle(45) is True
        assert result.all_angles(45) is True
        assert result.all_solid_lines() is True
        assert result.has_dashed_line() is False

    def test_analyse_checker(self):
        result = analyse("CHECKER")
        assert result.has_angle(0) is True
        assert result.has_angle(90) is True
        assert result.all_dashed_lines() is True

    def test_rotated_checker(self):
        pat = pattern.ISO_PATTERN["CHECKER"]
        result = pattern.PatternAnalyser(pattern.scale_pattern(pat, 2, 45))
        assert result.has_angle(45) is True
        assert result.has_angle(135) is True

    def test_analyse_crosses(self):
        result = analyse("CROSSES")
        assert result.has_angle(45) is True
        assert result.has_angle(135) is True
        assert result.all_dashed_lines() is True


@pytest.mark.parametrize(
    "angle,expected",
    [
        (0, 0),
        (45, 45),
        (90, 90),
        (135, 135),
        (22, 15),  # round to nearest 15° main angle
        (23, 30),  # round to nearest 15° main angle
        (45, 45),  # round to nearest 15° main angle
        (180, 0),  # only 1st and 2nd quadrant
        (270, 90),  # only 1st and 2nd quadrant
    ],
)
def test_round_angle_15_deg(angle, expected):
    assert pattern.round_angle_15_deg(angle) == expected


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import Union

ACAD_13 = "AC1012"
ACAD_14 = "AC1014"
ACAD_2000 = "AC1015"
ACAD_2004 = "AC1018"
ACAD_2007 = "AC1021"
ACAD_2010 = "AC1024"
ACAD_2013 = "AC1027"
ACAD_2018 = "AC1032"
ACAD_LATEST = ACAD_2018

SUPPORTED_VERSIONS = [ACAD_13, ACAD_14, ACAD_2000]
HEADER_ID = 0
CLASSES_ID = 1
OBJECTS_ID = 2
SENTINEL_SIZE = 16

Bytes = Union[bytes, bytearray, memoryview]


class DwgError(Exception):
    pass


class DwgVersionError(DwgError):
    pass


class DwgCorruptedFileHeader(DwgError):
    pass


class DwgCorruptedClassesSection(DwgError):
    pass


class DwgCorruptedHeaderSection(DwgError):
    pass


class CRCError(DwgError):
    pass



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
import pytest
import math

from ezdxf.path import (
    Path,
    make_path,
    converter,
    Command,
    tools,
)
from ezdxf.math import (
    Vec3,
    Vec2,
    Matrix44,
    Bezier4P,
    Bezier3P,
    close_vectors,
    OCS,
)
from ezdxf.entities import (
    factory,
    DXFEntity,
    Polymesh,
    LWPolyline,
    PolylinePath,
    EdgePath,
    Hatch,
)


def test_init():
    path = Path()
    assert path.start == (0, 0)
    assert len(path) == 0
    assert path.end == (0, 0)


def test_if_path_is_empty():
    path = Path()
    assert bool(path) is False, "should work in boolean tests"
    assert not len(path), "len() should work in boolean tests"
    assert len(path) == 0, "len() should be 0"


def test_if_path_is_not_empty():
    path = Path(start=(1, 0))
    path.line_to((2, 0))
    assert bool(path) is True, "should work in boolean tests"
    assert len(path), "len() should work boolean tests"
    assert len(path) > 0, "len() should be > 0"


def test_init_start():
    path = Path(start=(1, 2))
    assert path.start == (1, 2)


def test_if_path_with_only_a_start_point_is_still_empty():
    # Path() can not represent a point with any command
    path = Path(start=(1, 1))
    assert bool(path) is False, "should work in boolean tests"
    assert not len(path), "len() should work in boolean tests"
    assert len(path) == 0, "len() should be 0"


def test_line_to():
    path = Path()
    path.line_to((1, 2, 3))
    assert path[0] == (Vec3(1, 2, 3),)
    assert path.end == (1, 2, 3)


def test_path_requires_a_command_to_represent_a_point():
    path = Path((1, 1))
    path.line_to((1, 1))
    assert bool(path)
    assert len(path) > 0


def test_curve3_to():
    path = Path()
    path.curve3_to((10, 0), (5, 5))
    assert path[0] == ((10, 0), (5, 5))
    assert path.end == (10, 0)


def test_curve4_to():
    path = Path()
    path.curve4_to((1, 2, 3), (0, 1, 0), (0, 2, 0))
    assert path[0] == ((1, 2, 3), (0, 1, 0), (0, 2, 0))
    assert path.end == (1, 2, 3)


def test_user_data_is_none_by_default():
    assert Path().user_data is None


def test_set_and_get_user_data():
    path = Path()
    path.user_data = [1, 2, 3]
    assert path.user_data == [1, 2, 3]


def test_path_clones_share_user_data():
    path = Path()
    data = [1, 2, 3]
    path.user_data = data
    assert path.clone().user_data is data


def test_reversed_path_preserves_user_data():
    path = Path()
    path.user_data = "data"
    path.line_to((1, 2, 3))
    assert path.reversed().user_data == "data"


def test_transformed_path_preserves_user_data():
    path = Path()
    path.user_data = "data"
    path.line_to((1, 2, 3))
    assert path.transform(Matrix44()).user_data == "data"


def test_sub_paths_inherit_parent_user_data():
    path = Path()
    path.user_data = "data"
    path.line_to((1, 2, 3))
    path.move_to((7, 8, 9))
    path.line_to((7, 8, 9))
    assert path.has_sub_paths is True
    for p in path.sub_paths():
        assert p.user_data == "data"


def test_add_curves3():
    path = Path()
    c1 = Bezier3P(Vec2.list(((0, 0), (1, 1), (2, 0))))
    c2 = Bezier3P(Vec2.list(((2, 0), (1, -1), (0, 0))))
    tools.add_bezier3p(path, [c1, c2])
    assert len(path) == 2
    assert path.end == (0, 0)


def test_add_curves4():
    path = Path()
    c1 = Bezier4P(Vec2.list(((0, 0), (0, 1), (2, 1), (2, 0))))
    c2 = Bezier4P(Vec2.list(((2, 0), (2, -1), (0, -1), (0, 0))))
    tools.add_bezier4p(path, [c1, c2])
    assert len(path) == 2
    assert path.end == (0, 0)


def test_add_curves3_with_gap():
    path = Path()
    c1 = Bezier3P(Vec2.list(((0, 0), (1, 1), (2, 0))))
    c2 = Bezier3P(Vec2.list(((2, -1), (3, -2), (0, -1))))
    tools.add_bezier3p(path, [c1, c2])
    assert len(path) == 3  # added a line segment between curves
    assert path.end == (0, -1)


def test_add_curves4_with_gap():
    path = Path()
    c1 = Bezier4P(Vec3.list(((0, 0, 0), (0, 1, 0), (2, 1, 0), (2, 0, 0))))
    c2 = Bezier4P(Vec3.list(((2, -1, 0), (2, -2, 0), (0, -2, 0), (0, -1, 0))))
    tools.add_bezier4p(path, [c1, c2])
    assert len(path) == 3  # added a line segment between curves
    assert path.end == (0, -1, 0)


def test_add_curves3_reverse():
    path = Path(start=(0, 0))
    c1 = Bezier3P(Vec2.list(((2, 0), (1, 1), (0, 0))))
    tools.add_bezier3p(path, [c1])
    assert len(path) == 1
    assert path.end == (2, 0, 0)


def test_add_curves4_reverse():
    path = Path(start=(0, 0, 0))
    c1 = Bezier4P(Vec3.list(((2, 0, 0), (2, 1, 0), (0, 1, 0), (0, 0, 0))))
    tools.add_bezier4p(path, [c1])
    assert len(path) == 1
    assert path.end == (2, 0, 0)


class TestSubPath:
    def simple_multi_path(self):
        path = Path(start=(1, 0, 0))
        path.line_to((2, 0, 0))
        path.move_to((3, 0, 0))
        return path

    def test_has_no_sub_paths_by_default(self):
        path = Path()
        assert path.has_sub_paths is False

    def test_first_move_to(self):
        path = Path(start=(1, 0, 0))
        path.move_to((2, 0, 0))
        assert path.start.isclose((2, 0, 0)), "should reset the start point"
        assert len(path) == 0, "should not add a MOVETO cmd as first cmd"
        assert path.has_sub_paths is False

    def test_multiple_first_move_to(self):
        path = Path(start=(1, 0, 0))
        path.move_to((2, 0, 0))
        path.move_to((3, 0, 0))
        path.move_to((4, 0, 0))
        assert path.start.isclose((4, 0, 0)), "should reset the start point"
        assert len(path) == 0, "should not add a MOVETO cmd as first cmd"
        assert path.has_sub_paths is False

    def test_move_to_creates_a_multi_path_object(self):
        path = Path(start=(1, 0, 0))
        path.line_to((2, 0, 0))
        path.move_to((3, 0, 0))
        assert len(path) == 2, "should add a MOVETO cmd as last cmd"
        assert path.has_sub_paths is True, "should be a multi path object"
        assert path.end.isclose((3, 0, 0)), "should end at the MOVETO location"

    def test_merge_multiple_move_to_commands_at_the_end(self):
        path = self.simple_multi_path()
        path.move_to((4, 0, 0))
        path.move_to((4, 0, 0))
        assert (
            len(path) == 2
        ), "should merge multiple MOVETO commands at the end of the path"

    def test_clone_multi_path_object(self):
        path = self.simple_multi_path()
        path2 = path.clone()
        assert path2.has_sub_paths
        assert path.end == path2.end

    def test_cant_detect_orientation_of_multi_path_object(self):
        path = self.simple_multi_path()
        pytest.raises(TypeError, path.has_clockwise_orientation)

    def test_cant_convert_multi_path_object_to_clockwise_orientation(self):
        path = self.simple_multi_path()
        pytest.raises(TypeError, path.clockwise)

    def test_cant_convert_multi_path_object_to_ccw_orientation(self):
        path = self.simple_multi_path()
        pytest.raises(TypeError, path.counter_clockwise)

    def test_approximate_multi_path_object(self):
        path = self.simple_multi_path()
        vertices = list(path.approximate())
        assert len(vertices) == 3

    def test_flatten_multi_path_object(self):
        path = self.simple_multi_path()
        vertices = list(path.flattening(0.1))
        assert len(vertices) == 3

    def test_multi_path_object_to_wcs(self):
        path = self.simple_multi_path()
        path.to_wcs(OCS(), 0)
        assert path.end.isclose((3, 0, 0))

    def test_transform_multi_path_object(self):
        path = self.simple_multi_path()
        m = Matrix44.translate(1, 1, 1)
        path2 = path.transform(m)
        assert path.end.isclose((3, 0, 0))
        assert path2.has_sub_paths is True
        assert path2.end.isclose((4, 1, 1))

    def test_sub_paths_from_single_path_object(self):
        path = Path(start=(1, 2, 3))
        paths = list(path.sub_paths())
        assert len(paths) == 1
        s0 = paths[0]
        assert s0.start == (1, 2, 3)
        assert s0.end == (1, 2, 3)
        assert s0.has_sub_paths is False
        assert len(s0) == 0

    def test_sub_paths_from_multi_path_object(self):
        path = self.simple_multi_path()
        s0, s1 = path.sub_paths()
        assert s0.start == (1, 0, 0)
        assert s0.end == (2, 0, 0)
        assert s0.has_sub_paths is False
        assert len(s0) == 1

        assert s1.start == (3, 0, 0)
        assert s1.end == (3, 0, 0)
        assert len(s1) == 0
        assert s1.has_sub_paths is False


def test_add_spline():
    from ezdxf.math import BSpline

    spline = BSpline.from_fit_points([(2, 0), (4, 1), (6, -1), (8, 0)])
    path = Path()
    tools.add_spline(path, spline)
    assert path.start == (2, 0)
    assert path.end == (8, 0)

    # set start point to end of spline
    path = Path(start=(8, 0))
    # add reversed spline, by default the start of
    # an empty path is set to the spline start
    tools.add_spline(path, spline, reset=False)
    assert path.start == (8, 0)
    assert path.end == (2, 0)

    path = Path()
    # add a line segment from (0, 0) to start of spline
    tools.add_spline(path, spline, reset=False)
    assert path.start == (0, 0)
    assert path.end == (8, 0)


def test_from_spline():
    spline = factory.new("SPLINE")
    spline.fit_points = [(2, 0), (4, 1), (6, -1), (8, 0)]
    path = make_path(spline)
    assert path.start.isclose((2, 0))
    assert path.end.isclose((8, 0))


def test_add_ellipse():
    from ezdxf.math import ConstructionEllipse

    ellipse = ConstructionEllipse(
        center=(3, 0),
        major_axis=(1, 0),
        ratio=0.5,
        start_param=0,
        end_param=math.pi,
    )
    path = Path()
    tools.add_ellipse(path, ellipse)
    assert path.start.isclose((4, 0))
    assert path.end.isclose((2, 0))

    # set start point to end of ellipse
    path = Path(start=(2, 0))
    # add reversed ellipse, by default the start of
    # an empty path is set to the ellipse start
    tools.add_ellipse(path, ellipse, reset=False)
    assert path.start.isclose((2, 0))
    assert path.end.isclose((4, 0))

    path = Path()
    # add a line segment from (0, 0) to start of ellipse
    tools.add_ellipse(path, ellipse, reset=False)
    assert path.start.isclose((0, 0))
    assert path.end.isclose((2, 0))


def test_raises_type_error_for_unsupported_objects():
    with pytest.raises(TypeError):
        make_path(DXFEntity())
    with pytest.raises(TypeError):
        make_path(Polymesh.new(dxfattribs={"flags": Polymesh.POLYMESH}))
    with pytest.raises(TypeError):
        make_path(Polymesh.new(dxfattribs={"flags": Polymesh.POLYFACE}))


def test_from_ellipse():
    ellipse = factory.new(
        "ELLIPSE",
        dxfattribs={
            "center": (3, 0),
            "major_axis": (1, 0),
            "ratio": 0.5,
            "start_param": 0,
            "end_param": math.pi,
        },
    )
    path = make_path(ellipse)
    assert path.start.isclose((4, 0))
    assert path.end.isclose((2, 0))


def test_from_arc():
    arc = factory.new(
        "ARC",
        dxfattribs={
            "center": (1, 0, 0),
            "radius": 1,
            "start_angle": 0,
            "end_angle": 180,
        },
    )
    path = make_path(arc)
    assert path.start.isclose((2, 0))
    assert path.end.isclose((0, 0))


@pytest.mark.parametrize("radius", [1, -1])
def test_from_circle(radius):
    circle = factory.new(
        "CIRCLE",
        dxfattribs={
            "center": (1, 0, 0),
            "radius": radius,
        },
    )
    path = make_path(circle)
    assert path.start.isclose((2, 0))
    assert path.end.isclose((2, 0))
    assert path.is_closed is True


def test_from_circle_with_zero_radius():
    circle = factory.new(
        "CIRCLE",
        dxfattribs={
            "center": (1, 0, 0),
            "radius": 0,
        },
    )
    path = make_path(circle)
    assert len(path) == 0


def test_from_line():
    start = Vec3(1, 2, 3)
    end = Vec3(4, 5, 6)
    line = factory.new("LINE", dxfattribs={"start": start, "end": end})
    path = make_path(line)
    assert path.start.isclose(start)
    assert path.end.isclose(end)


@pytest.mark.parametrize("dxftype", ["SOLID", "TRACE", "3DFACE"])
def test_from_quadrilateral_with_4_points(dxftype):
    entity = factory.new(dxftype)
    entity.dxf.vtx0 = (0, 0, 0)
    entity.dxf.vtx1 = (1, 0, 0)
    entity.dxf.vtx2 = (1, 1, 0)
    entity.dxf.vtx3 = (0, 1, 0)
    path = make_path(entity)
    assert path.start == (0, 0, 0)
    assert path.is_closed is True
    assert len(list(path.approximate())) == 5


@pytest.mark.parametrize("dxftype", ["SOLID", "TRACE", "3DFACE"])
def test_from_quadrilateral_with_3_points(dxftype):
    entity = factory.new(dxftype)
    entity.dxf.vtx0 = (0, 0, 0)
    entity.dxf.vtx1 = (1, 0, 0)
    entity.dxf.vtx2 = (1, 1, 0)
    entity.dxf.vtx3 = (1, 1, 0)  # last two points are equal
    path = make_path(entity)
    assert path.is_closed is True
    assert len(list(path.approximate())) == 4


def test_lwpolyline_lines():
    from ezdxf.entities import LWPolyline

    pline = LWPolyline()
    pline.append_points([(1, 1), (2, 1), (2, 2)], format="xy")
    path = make_path(pline)
    assert path.start.isclose((1, 1))
    assert path.end.isclose((2, 2))
    assert len(path) == 2

    pline.dxf.elevation = 1.0
    path = make_path(pline)
    assert path.start.isclose((1, 1, 1))
    assert path.end.isclose((2, 2, 1))


POINTS = [
    (0, 0, 0),
    (3, 0, -1),
    (6, 0, 0),
    (9, 0, 0),
    (9, -3, 0),
]


def test_make_path_from_lwpolyline_with_bulges():
    pline = LWPolyline()
    pline.closed = True
    pline.append_points(POINTS, format="xyb")
    path = make_path(pline)
    assert path.start == (0, 0)
    assert path.end == (0, 0)  # closed
    assert any(cmd.type == Command.CURVE4_TO for cmd in path)


def test_make_path_from_full_circle_lwpolyline():
    pline = LWPolyline()
    pline.closed = True
    pline.append_points([(0, 0, 1), (1, 0, 1)], format="xyb")
    path = make_path(pline)
    assert path.start.isclose((0, 0))
    assert path.end.isclose((0, 0))
    assert len(path) == 4
    assert any(cmd.type == Command.CURVE4_TO for cmd in path)
    vertices = list(path.flattening(0.1, segments=16))
    assert len(vertices) == 65


def test_make_path_from_full_circle_lwpolyline_issue_424():
    pline = LWPolyline()
    pline.closed = True
    points = [
        (39_482_129.9462793, 3_554_328.753243976, 1.0),
        (39_482_129.95781776, 3_554_328.753243976, 1.0),
    ]
    pline.append_points(points, format="xyb")
    path = make_path(pline)
    assert len(path) == 2


S_SHAPE = [
    (0, 0, 0),
    (5, 0, 1),
    (5, 1, 0),
    (0, 1, -1),
    (0, 2, 0),
    (5, 2, 0),
]


def test_lwpolyline_s_shape():
    from ezdxf.entities import LWPolyline

    pline = LWPolyline()
    pline.append_points(S_SHAPE, format="xyb")
    path = make_path(pline)
    assert path.start == (0, 0)
    assert path.end == (5, 2)  # closed
    assert any(cmd.type == Command.CURVE4_TO for cmd in path)


def test_polyline_lines():
    from ezdxf.entities import Polyline

    pline = Polyline()
    pline.append_formatted_vertices([(1, 1), (2, 1), (2, 2)], format="xy")
    path = make_path(pline)
    assert path.start == (1, 1)
    assert path.end == (2, 2)
    assert len(path) == 2

    pline.dxf.elevation = (0, 0, 1)
    path = make_path(pline)
    assert path.start == (1, 1, 1)
    assert path.end == (2, 2, 1)


def test_polyline_with_bulges():
    from ezdxf.entities import Polyline

    pline = Polyline()
    pline.close(True)
    pline.append_formatted_vertices(POINTS, format="xyb")
    path = make_path(pline)
    assert path.start == (0, 0)
    assert path.end == (0, 0)  # closed
    assert any(cmd.type == Command.CURVE4_TO for cmd in path)


def test_3d_polyline():
    from ezdxf.entities import Polyline

    pline = Polyline.new(dxfattribs={"flags": Polyline.POLYLINE_3D})
    pline.append_vertices([(1, 1, 1), (2, 1, 3), (2, 2, 2)])
    path = make_path(pline)
    assert path.start == (1, 1, 1)
    assert path.end == (2, 2, 2)
    assert len(path) == 2


POLYLINE_POINTS = [
    # x, y, b
    (0, 0, 0),
    (2, 2, -1),
    (4, 0, 1),
    (6, 0, 0),
]


class TestPathFromBoundaryWithElevationAndFlippedExtrusion:
    @pytest.fixture
    def hatch(self):
        return Hatch.new(
            dxfattribs={
                "elevation": (0, 0, 4),
                "extrusion": (0, 0, -1),
            }
        )

    def test_from_hatch_polyline_path(self, hatch):
        hatch.paths.add_polyline_path(POLYLINE_POINTS)
        path = make_path(hatch)
        assert path.has_curves is True
        assert len(path) > 5
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())


def test_approximate_lines():
    path = Path()
    path.line_to((1, 1))
    path.line_to((2, 0))
    vertices = list(path.approximate())
    assert len(vertices) == 3
    assert vertices[0] == path.start == (0, 0)
    assert vertices[2] == path.end == (2, 0)


def test_approximate_curves():
    path = Path()
    path.curve3_to((2, 0), (1, 1))
    path.curve4_to((3, 0), (2, 1), (3, 1))
    vertices = list(path.approximate(20))
    assert len(vertices) == 41
    assert vertices[0] == (0, 0)
    assert vertices[-1] == (3, 0)


def test_path_from_hatch_polyline_path_without_bulge():
    polyline_path = PolylinePath()
    polyline_path.set_vertices([(0, 0), (0, 1), (1, 1), (1, 0)], is_closed=False)
    path = converter.from_hatch_polyline_path(polyline_path)
    assert len(path) == 3
    assert path.start == (0, 0)
    assert path.end == (1, 0)

    polyline_path.is_closed = True
    path = converter.from_hatch_polyline_path(polyline_path)
    assert len(path) == 4
    assert path.start == (0, 0)
    assert path.end == (0, 0)


def test_path_from_hatch_polyline_path_with_bulge():
    polyline_path = PolylinePath()
    polyline_path.set_vertices([(0, 0), (1, 0, 0.5), (2, 0), (3, 0)], is_closed=False)
    path = converter.from_hatch_polyline_path(polyline_path)
    assert len(path) == 4
    assert path.start == (0, 0)
    assert path.end == (3, 0)

    assert path[1].type == Command.CURVE4_TO
    assert path[1].end.isclose((1.5, -0.25))


@pytest.fixture
def p1():
    path = Path()
    path.line_to((2, 0))
    path.curve4_to((4, 0), (2, 1), (4, 1))  # end, ctrl1, ctrl2
    path.curve3_to((6, 0), (5, -1))  # end, ctrl
    return path


def test_path_cloning(p1):
    p2 = p1.clone()
    for cmd1, cmd2 in zip(p1, p2):
        assert cmd1 == cmd2

    # but have different command lists:
    p2.line_to((4, 4))
    assert len(p2) == len(p1) + 1


def test_approximate_line_curves(p1):
    vertices = list(p1.approximate(20))
    assert len(vertices) == 42
    assert vertices[0].isclose((0, 0))
    assert vertices[-1].isclose((6, 0))


def test_transform(p1):
    p2 = p1.transform(Matrix44.translate(1, 1, 0))
    assert p2.start.isclose((1, 1))
    assert p2[0].end.isclose((3, 1))  # line to location
    assert p2[1].end.isclose((5, 1))  # cubic to location
    assert p2[1].ctrl1.isclose((3, 2))  # cubic ctrl1
    assert p2[1].ctrl2.isclose((5, 2))  # cubic ctrl2
    assert p2[2].end.isclose((7, 1))  # quadratic to location
    assert p2[2].ctrl.isclose((6, 0))  # quadratic ctrl
    assert p2.end.isclose((7, 1))


def test_control_vertices(p1):
    vertices = list(p1.control_vertices())
    assert close_vectors(
        vertices, [(0, 0), (2, 0), (2, 1), (4, 1), (4, 0), (5, -1), (6, 0)]
    )
    path = Path()
    assert len(list(path.control_vertices())) == 0
    assert list(path.control_vertices()) == list(path.approximate(2))
    path = converter.from_vertices([(0, 0), (1, 0)])
    assert len(list(path.control_vertices())) == 2


def test_has_clockwise_orientation():
    # basic has_clockwise_orientation() function is tested in:
    # test_617_clockwise_orientation
    path = converter.from_vertices([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert path.has_clockwise_orientation() is False

    path = Path()
    path.line_to((2, 0))
    path.curve4_to((4, 0), (2, 1), (4, 1))  # end, ctrl1, ctrl2
    assert path.has_clockwise_orientation() is True


class TestReversePath:
    def test_reversing_empty_path(self):
        p = Path()
        assert len(p.reversed()) == 0

    def test_reversing_one_line(self):
        p = Path()
        p.line_to((1, 0))
        p2 = list(p.reversed().control_vertices())
        assert close_vectors(p2, [(1, 0), (0, 0)])

    def test_reversing_one_curve3(self):
        p = Path()
        p.curve3_to((3, 0), (1.5, 1))
        p2 = list(p.reversed().control_vertices())
        assert close_vectors(p2, [(3, 0), (1.5, 1), (0, 0)])

    def test_reversing_one_curve4(self):
        p = Path()
        p.curve4_to((3, 0), (1, 1), (2, 1))
        p2 = list(p.reversed().control_vertices())
        assert close_vectors(p2, [(3, 0), (2, 1), (1, 1), (0, 0)])

    def test_reversing_path_ctrl_vertices(self, p1):
        p2 = p1.reversed()
        assert close_vectors(
            p2.control_vertices(), reversed(list(p1.control_vertices()))
        )

    def test_reversing_path_approx(self, p1):
        p2 = p1.reversed()
        v1 = list(p1.approximate())
        v2 = list(p2.approximate())
        assert close_vectors(v1, reversed(v2))

    def test_reversing_multi_path(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        p.line_to((3, 0, 0))
        r = p.reversed()
        assert r.has_sub_paths is True
        assert len(r) == 3
        assert r.start == (3, 0, 0)
        assert r.end == (0, 0, 0)

        r0, r1 = r.sub_paths()
        assert r0.start == (3, 0, 0)
        assert r0.end == (2, 0, 0)
        assert r1.start == (1, 0, 0)
        assert r1.end == (0, 0, 0)

    def test_reversing_multi_path_with_a_move_to_cmd_at_the_end(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        # The last move_to will become the first move_to.
        # A move_to as first command just moves the start point.
        r = p.reversed()
        assert len(r) == 1
        assert r.start == (1, 0, 0)
        assert r.end == (0, 0, 0)
        assert r.has_sub_paths is False


def test_cw_and_ccw_orientation(p1):
    from ezdxf.math import has_clockwise_orientation

    cw_path = p1.clockwise()
    ccw_path = p1.counter_clockwise()
    assert has_clockwise_orientation(cw_path.control_vertices()) is True
    assert has_clockwise_orientation(ccw_path.control_vertices()) is False


@pytest.fixture
def edge_path():
    ep = EdgePath()
    ep.add_line(
        (70.79594401862802, 38.81021154906707),
        (61.49705431814723, 38.81021154906707),
    )
    ep.add_ellipse(
        center=(49.64089977339618, 36.43095770602131),
        major_axis=(16.69099826506408, 6.96203799241026),
        ratio=0.173450304570581,
        start_angle=348.7055398636587,
        end_angle=472.8737032507014,
        ccw=True,
    )
    ep.add_line(
        (47.21845383585098, 38.81021154906707),
        (32.00406637283394, 38.81021154906707),
    )
    ep.add_arc(
        center=(27.23255482392775, 37.32841621274949),
        radius=4.996302620946588,
        start_angle=17.25220809399113,
        end_angle=162.7477919060089,
        ccw=True,
    )
    ep.add_line(
        (22.46104327502155, 38.81021154906707),
        (15.94617981131185, 38.81021154906707),
    )
    ep.add_line(
        (15.94617981131185, 38.81021154906707),
        (15.94617981131185, 17.88970141145027),
    )
    ep.add_line(
        (15.94617981131185, 17.88970141145027),
        (22.07965616927404, 17.88970141145026),
    )
    ep.add_spline(
        control_points=[
            (22.07965616927404, 17.88970141145027),
            (23.44151487263461, 19.56130038573538),
            (28.24116384863678, 24.26061858002495),
            (35.32501805918895, 14.41241846270862),
            (46.6153937930182, 11.75667640124574),
            (47.53794331191931, 23.11460620899234),
            (51.8076764251228, 12.06821526039212),
            (60.37405963053161, 14.60131364832752),
            (63.71393926002737, 20.24075830571701),
            (67.36423789268184, 19.07462271006858),
            (68.72358721334537, 17.88970141145026),
        ],
        knot_values=[
            2.825276861104652,
            2.825276861104652,
            2.825276861104652,
            2.825276861104652,
            8.585563484895022,
            22.93271064560279,
            29.77376253023298,
            35.89697937194972,
            41.26107011625705,
            51.23489795733507,
            54.82267350174899,
            59.57512798605262,
            59.57512798605262,
            59.57512798605262,
            59.57512798605262,
        ],
        degree=3,
        periodic=0,
    )
    ep.add_line(
        (68.72358721334535, 17.88970141145027),
        (70.79594401862802, 17.88970141145027),
    )
    ep.add_line(
        (70.79594401862802, 17.88970141145027),
        (70.79594401862802, 38.81021154906707),
    )
    return ep


def test_from_edge_path(edge_path):
    path = converter.from_hatch_edge_path(edge_path)
    assert path.has_sub_paths is False
    assert len(path) == 19


def test_from_edge_path_with_two_closed_loops():
    ep = EdgePath()
    # 1st loop: closed segments
    ep.add_line((0, 0), (0, 1))
    ep.add_line((0, 1), (1, 1))
    ep.add_line((1, 1), (0, 1))
    ep.add_line((0, 1), (0, 0))

    # 2nd loop: closed segments
    ep.add_line((2, 0), (3, 0))
    ep.add_line((3, 0), (3, 1))
    ep.add_line((3, 1), (2, 1))
    ep.add_line((2, 1), (2, 0))
    path = converter.from_hatch_edge_path(ep)
    assert path.has_sub_paths is True, "should return a multi-path"
    assert len(list(path.sub_paths())) == 2, "expected two sub paths"


LOOP = Vec2(0, 0), Vec2(1, 0), Vec2(1, 1), Vec2(0, 1)
A, B, C, D = LOOP


def test_edge_path_loops_with_gaps_should_be_closed():
    # behavior based on issue #706
    ep = EdgePath()
    ep.add_line(A, B)  # gap B -> C
    ep.add_line(C, D)  # gap D -> A
    path = converter.from_hatch_edge_path(ep)
    assert len(path) == 4
    assert path.is_closed is True, "expected a closed loop"


@pytest.mark.parametrize(
    "e0,e1,e2,e3",
    [
        [(A, B), (B, C), (C, D), (D, A)],  # case 0: consecutive order
        #    0---->  1---->  2---->        # end - start
        # <--------------------------3
        [(D, C), (C, B), (B, A), (A, D)],  # case 1: reversed order
        [(A, B), (C, B), (C, D), (D, A)],  # case 2
        #    0------->       2---->        # 0: end - end, reversing (C, B)
        #         1------->                # 1: end - start
        # <--------------------------3
        [(A, B), (C, B), (D, C), (D, A)],  # case 3
        #    0------->    2------->        # 0: end - end, reversing (C, B)
        #         1---------->             # 1: end - end
        # <--------------------------3
        [(A, B), (D, A), (D, C), (B, C)],  # case 4
        # 0---------->       2------->     # 0: start - end; 2: end - end, rev: B, C
        #         1------->                # 1: start - start
        #    <--------------------3
        [(A, B), (D, A), (D, C), (C, B)],  # case 5
        # 0---------->       2---->        # 0: start - end
        #         1------->                # 1: start - start
        #    <-----------------------3
        [(A, B), (D, A), (C, D), (B, C)],  # case 6
        # 0---------->    2---------->     # 0: start - end
        #         1---------->             # 1: start - end
        #    <--------------------3
        [(A, B), (B, C), (A, D), (C, D)],  # case 7
        #    0---->          2------->     # 0: start - end
        # <---------------1                # 1: start - start
        #            <------------3
    ],
)
def test_edge_path_closed_loop(e0, e1, e2, e3):
    ep = EdgePath()
    ep.add_line(e0[0], e0[1])
    ep.add_line(e1[0], e1[1])
    ep.add_line(e2[0], e2[1])
    ep.add_line(e3[0], e3[1])
    path = converter.from_hatch_edge_path(ep)
    assert len(list(path.sub_paths())) == 1, "expected one closed loop"
    assert len(list(path.control_vertices())) == 5
    assert path.is_closed is True, "expected a closed loop"


class TestPathFromEdgePathWithElevationAndFlippedExtrusion:
    def test_line_edge(self):
        ep = EdgePath()
        ep.add_line(A, B)
        ep.add_line(B, C)
        ep.add_line(C, D)
        ep.add_line(D, A)
        path = converter.from_hatch_edge_path(ep, ocs=OCS((0, 0, -1)), elevation=4)
        assert len(list(path.sub_paths())) == 1, "expected one closed loop"
        assert len(list(path.control_vertices())) == 5
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())
        assert path.is_closed is True, "expected a closed loop"

    def test_arc_edge(self):
        ep = EdgePath()
        ep.add_arc(
            center=(5.0, 5.0),
            radius=5.0,
            start_angle=0,
            end_angle=90,
            ccw=True,
        )
        ep.add_line((5, 10), (10, 5))
        path = converter.from_hatch_edge_path(ep, ocs=OCS((0, 0, -1)), elevation=4)
        assert len(path) == 2
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())

    def test_ellipse_edge(self):
        ep = EdgePath()
        ep.add_ellipse(
            center=(5.0, 5.0),
            major_axis=(5.0, 0.0),
            ratio=1,
            start_angle=0,
            end_angle=90,
            ccw=True,
        )
        ep.add_line((5, 10), (10, 5))
        path = converter.from_hatch_edge_path(ep, ocs=OCS((0, 0, -1)), elevation=4)
        assert len(path) == 2
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())

    def test_spline_edge(self):
        ep = EdgePath()
        ep.add_spline(fit_points=[(10, 5), (8, 5), (6, 8), (5, 10)])
        ep.add_line((5, 10), (10, 5))
        path = converter.from_hatch_edge_path(ep, ocs=OCS((0, 0, -1)), elevation=4)
        assert len(path) > 2
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())

    def test_from_complex_edge_path(self, edge_path):
        path = converter.from_hatch_edge_path(
            edge_path, ocs=OCS((0, 0, -1)), elevation=4
        )
        assert path.has_sub_paths is False
        assert len(path) == 19
        assert all(math.isclose(v.z, -4) for v in path.control_vertices())


def test_extend_path_by_another_none_empty_path():
    p0 = Path((1, 0, 0))
    p0.line_to((2, 0, 0))
    p1 = Path((3, 0, 0))
    p1.line_to((3, 0, 0))
    p0.extend_multi_path(p1)
    assert p0.has_sub_paths is True
    assert p0.start == (1, 0, 0)
    assert p0.end == (3, 0, 0)


def test_extend_path_by_another_single_path():
    path = Path((1, 0, 0))
    path.line_to((2, 0, 0))
    p1 = Path((3, 0, 0))
    p1.line_to((4, 0, 0))
    path.extend_multi_path(p1)
    assert path.has_sub_paths is True
    assert path.start == (1, 0, 0)
    assert path.end == (4, 0, 0)


def test_extend_path_by_another_multi_path():
    path = Path((1, 0, 0))
    path.line_to((2, 0, 0))
    p1 = Path((3, 0, 0))
    p1.line_to((4, 0, 0))
    p1.move_to((5, 0, 0))
    path.extend_multi_path(p1)
    assert path.has_sub_paths is True
    assert path.start == (1, 0, 0)
    assert path.end == (5, 0, 0)


def test_append_empty_path():
    path = Path((1, 0, 0))
    path.line_to((2, 0, 0))
    start = path.start
    end = path.end
    path.append_path(Path())
    assert start == path.start and end == path.end, "path should be unchanged"


def test_append_path_without_a_gap():
    p1 = Path((1, 0, 0))
    p1.line_to((2, 0, 0))
    p2 = Path((2, 0, 0))
    p2.line_to((3, 0, 0))
    p1.append_path(p2)
    assert p1.start == (1, 0)
    assert p1.end == (3, 0)
    assert len(p1) == 2


def test_append_path_with_a_gap():
    p1 = Path((1, 0, 0))
    p1.line_to((2, 0, 0))
    p2 = Path((3, 0, 0))
    p2.line_to((4, 0, 0))
    p1.append_path(p2)
    assert p1.start == (1, 0)
    assert p1.end == (4, 0)
    assert len(p1) == 3


class TestCloseSubPath:
    def test_close_last_sub_path(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        p.line_to((3, 0, 0))
        p.close_sub_path()
        assert p.end == (2, 0, 0)

    def test_does_nothing_if_last_sub_path_is_closed(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        p.line_to((3, 0, 0))
        p.line_to((2, 0, 0))
        assert len(p) == 4
        p.close_sub_path()
        assert len(p) == 4
        assert p.end == (2, 0, 0)

    def test_does_nothing_if_last_sub_path_is_empty(self):
        p = Path()
        p.line_to((1, 0, 0))
        p.move_to((2, 0, 0))
        assert len(p) == 2
        p.close_sub_path()
        assert len(p) == 2
        assert p.end == (2, 0, 0)

    def test_close_single_path(self):
        p = Path((1, 0, 0))
        p.line_to((3, 0, 0))
        p.close_sub_path()
        assert p.end == (1, 0, 0)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence, no_type_check, Any, Callable, Dict, List, Tuple
from typing_extensions import TypeAlias, override
import abc
import json

from ezdxf.math import Vec2, world_mercator_to_gps
from ezdxf.path import Command, nesting
from ezdxf.npshapes import orient_paths, single_paths

from .type_hints import Color
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import Configuration
from .properties import BackendProperties


__all__ = ["CustomJSONBackend", "GeoJSONBackend"]

CUSTOM_JSON_SPECS = """
JSON content = [entity, entity, ...]

entity = {
    "type": point | lines | path | filled-paths | filled-polygon,
    "properties": {
        "color": "#RRGGBBAA",
        "stroke-width": 0.25, # in mm
        "layer": "name"
    },
    "geometry": depends on "type"
}
DXF linetypes (DASH, DOT, ...) are resolved into solid lines.

A single point:
point = {
    "type": "point",
    "properties": {...},
    "geometry": [x, y]
}

Multiple lines with common properties:
lines = {
    "type": "lines",
    "properties": {...},
    "geometry": [
        (x0, y0, x1, y1),  # 1. line
        (x0, y0, x1, y1),  # 2. line
        ....
    ]
}
Lines can contain points where x0 == x1 and y0 == y1!

A single linear path without filling:
path = {
    "type": "path",
    "properties": {...},
    "geometry": [path-command, ...]
}

SVG-like path structure:
- The first path-command is always an absolute move to "M"
- The "M" command does not appear inside a path, each path is a continuouse geometry 
  (no multi-paths).

path-command = 
    ("M", x, y) = absolute move to
    ("L", x, y) = absolute line to
    ("Q", x0, y0, x1, y1) = absolute quadratice Bezier curve to
    - (x0, y0) = control point
    - (x1, y1) = end point
    ("C", x0, y0, x1, y1, x2, y2) = absolute cubic Bezier curve to
    - (x0, y0) = control point 1
    - (x1, y1) = control point 2
    - (x2, y2) = end point
    ("Z",) = close path

Multiple filled paths:

Exterior paths and holes are mixed and NOT oriented by default (clockwise or 
counter-clockwise) - PyQt and SVG have no problem with that structure but matplotlib 
requires oriented paths.  When oriented paths are required the CustomJSONBackend can 
orient the paths on demand.

filled-paths = {
    "type": "filled-paths",
    "properties": {...},
    "geometry": [
        [path-command, ...],  # 1. path
        [path-command, ...],  # 2. path  
        ...
    ]
}

A single filled polygon:
A polygon is explicitly closed, so first vertex == last vertex is guaranteed.
filled-polygon = {
    "type": "filled-polygon",
    "properties": {...},
    "geometry": [
        (x0, y0),
        (x1, y1),
        (x2, y2),
        ...
    ]
}

"""


class _JSONBackend(BackendInterface):
    def __init__(self) -> None:
        self._entities: list[dict[str, Any]] = []
        self.max_sagitta = 0.01  # set by configure()
        self.min_lineweight = 0.05  # in mm, set by configure()
        self.lineweight_scaling = 1.0  # set by configure()
        # set fixed lineweight for all strokes:
        # set Configuration.min_lineweight to the desired lineweight in 1/300 inch!
        # set Configuration.lineweight_scaling to 0
        self.fixed_lineweight = 0.0

    @abc.abstractmethod
    def get_json_data(self) -> Any: ...
    def get_string(self, *, indent: int | str = 2) -> str:
        """Returns the result as a JSON string."""
        return json.dumps(self.get_json_data(), indent=indent)

    @override
    def configure(self, config: Configuration) -> None:
        if config.min_lineweight:
            # config.min_lineweight in 1/300 inch!
            min_lineweight_mm = config.min_lineweight * 25.4 / 300
            self.min_lineweight = max(0.05, min_lineweight_mm)
        self.lineweight_scaling = config.lineweight_scaling
        if self.lineweight_scaling == 0.0:
            # use a fixed lineweight for all strokes defined by min_lineweight
            self.fixed_lineweight = self.min_lineweight
        self.max_sagitta = config.max_flattening_distance

    @override
    def clear(self) -> None:
        self._entities.clear()

    @override
    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        pass

    @override
    def set_background(self, color: Color) -> None:
        pass

    @override
    def finalize(self) -> None:
        pass

    @override
    def enter_entity(self, entity, properties) -> None:
        pass

    @override
    def exit_entity(self, entity) -> None:
        pass


MOVE_TO_ABS = "M"
LINE_TO_ABS = "L"
QUAD_TO_ABS = "Q"
CUBIC_TO_ABS = "C"
CLOSE_PATH = "Z"


class CustomJSONBackend(_JSONBackend):
    """Creates a JSON-like output with a custom JSON scheme.  This scheme supports
    curved shapes by a SVG-path like structure and coordinates are not limited in
    any way.  This backend can be used to send geometries from a web-backend to a
    frontend.

    The JSON scheme is documented in the source code:

    https://github.com/mozman/ezdxf/blob/master/src/ezdxf/addons/drawing/json.py

    Args:
        orient_paths: orient exterior and hole paths on demand, exterior paths have
            counter-clockwise orientation and holes have clockwise orientation.

    **Class Methods**

    .. automethod:: get_json_data

    .. automethod:: get_string

    .. versionadded:: 1.3.0

    """

    def __init__(self, orient_paths=False) -> None:
        super().__init__()
        self.orient_paths = orient_paths

    @override
    def get_json_data(self) -> list[dict[str, Any]]:
        """Returns the result as a JSON-like data structure."""
        return self._entities

    def add_entity(
        self, entity_type: str, geometry: Sequence[Any], properties: BackendProperties
    ):
        if not geometry:
            return
        self._entities.append(
            {
                "type": entity_type,
                "properties": self.make_properties_dict(properties),
                "geometry": geometry,
            }
        )

    def make_properties_dict(self, properties: BackendProperties) -> dict[str, Any]:
        if self.fixed_lineweight:
            stroke_width = self.fixed_lineweight
        else:
            stroke_width = max(
                self.min_lineweight, properties.lineweight * self.lineweight_scaling
            )
        return {
            "color": properties.color,
            "stroke-width": round(stroke_width, 2),
            "layer": properties.layer,
        }

    @override
    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.add_entity("point", [pos.x, pos.y], properties)

    @override
    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        self.add_entity("lines", [(start.x, start.y, end.x, end.y)], properties)

    @override
    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        lines = list(lines)
        if len(lines) == 0:
            return
        self.add_entity("lines", [(s.x, s.y, e.x, e.y) for s, e in lines], properties)

    @override
    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        self.add_entity("path", make_json_path(path), properties)

    @override
    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        paths = list(paths)
        if len(paths) == 0:
            return
        if self.orient_paths:
            paths = orient_paths(paths)  # returns single paths
        else:
            # Just single paths allowed, no multi paths!
            paths = single_paths(paths)
        json_paths: list[Any] = []
        for path in paths:
            if len(path):
                json_paths.append(make_json_path(path, close=True))
        if json_paths:
            self.add_entity("filled-paths", json_paths, properties)

    @override
    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        vertices: list[Vec2] = points.vertices()
        if len(vertices) < 3:
            return
        if not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        self.add_entity("filled-polygon", [(v.x, v.y) for v in vertices], properties)


@no_type_check
def make_json_path(path: BkPath2d, close=False) -> list[Any]:
    if len(path) == 0:
        return []
    end: Vec2 = path.start
    commands: list = [(MOVE_TO_ABS, end.x, end.y)]
    for cmd in path.commands():
        end = cmd.end
        if cmd.type == Command.MOVE_TO:
            commands.append((MOVE_TO_ABS, end.x, end.y))
        elif cmd.type == Command.LINE_TO:
            commands.append((LINE_TO_ABS, end.x, end.y))
        elif cmd.type == Command.CURVE3_TO:
            c1 = cmd.ctrl
            commands.append((QUAD_TO_ABS, c1.x, c1.y, end.x, end.y))
        elif cmd.type == Command.CURVE4_TO:
            c1 = cmd.ctrl1
            c2 = cmd.ctrl2
            commands.append((CUBIC_TO_ABS, c1.x, c1.y, c2.x, c2.y, end.x, end.y))
    if close:
        commands.append(CLOSE_PATH)
    return commands


# dict and list not allowed here for Python < 3.10
PropertiesMaker: TypeAlias = Callable[[str, float, str], Dict[str, Any]]
TransformFunc: TypeAlias = Callable[[Vec2], Tuple[float, float]]
# GeoJSON ring
Ring: TypeAlias = List[Tuple[float, float]]

# The first ring is the exterior path followed by optional holes, nested polygons are
# not supported by the GeoJSON specification.
GeoJsonPolygon: TypeAlias = List[Ring]


def properties_maker(color: str, stroke_width: float, layer: str) -> dict[str, Any]:
    """Returns the property dict::

        {
            "color": color,
            "stroke-width": stroke_width,
            "layer": layer,
        }

    Returning an empty dict prevents properties in the GeoJSON output and also avoids
    wraping entities into "Feature" objects.
    """
    return {
        "color": color,
        "stroke-width": round(stroke_width, 2),
        "layer": layer,
    }


def no_transform(location: Vec2) -> tuple[float, float]:
    """Dummy transformation function.  Does not apply any transformations and
    just returns the input coordinates.
    """
    return (location.x, location.y)


def make_world_mercator_to_gps_function(tol: float = 1e-6) -> TransformFunc:
    """Returns a function to transform WGS84 World Mercator `EPSG:3395 <https://epsg.io/3395>`_
    location given as cartesian 2D coordinates x, y in meters into WGS84 decimal
    degrees as longitude and latitude `EPSG:4326 <https://epsg.io/4326>`_ as
    used by GPS.

    Args:
        tol: accuracy for latitude calculation

    """

    def _transform(location: Vec2) -> tuple[float, float]:
        """Transforms WGS84 World Mercator EPSG:3395 coordinates to WGS84 EPSG:4326."""
        return world_mercator_to_gps(location.x, location.y, tol)
    return _transform


class GeoJSONBackend(_JSONBackend):
    """Creates a JSON-like output according the `GeoJSON`_ scheme.
    GeoJSON uses a geographic coordinate reference system, World Geodetic
    System 1984 `EPSG:4326 <https://epsg.io/4326>`_, and units of decimal degrees.

    - Latitude: -90 to +90 (South/North)
    - Longitude: -180 to +180 (East/West)

    So most DXF files will produce invalid coordinates and it is the job of the
    **package-user** to provide a function to transfrom the input coordinates to
    EPSG:4326!  The :class:`~ezdxf.addons.drawing.recorder.Recorder` and
    :class:`~ezdxf.addons.drawing.recorder.Player` classes can help to detect the
    extents of the DXF content.

    Default implementation:

    .. autofunction:: no_transform

    Factory function to make a transform function from WGS84 World Mercator
    `EPSG:3395 <https://epsg.io/3395>`_  coordinates to WGS84 (GPS) 
    `EPSG:4326 <https://epsg.io/4326>`_.

    .. autofunction:: make_world_mercator_to_gps_function

    The GeoJSON format supports only straight lines so curved shapes are flattened to
    polylines and polygons.

    The properties are handled as a foreign member feature and is therefore not defined
    in the GeoJSON specs.  It is possible to provide a custom function to create these
    property objects.

    Default implementation:

    .. autofunction:: properties_maker


    Args:
        properties_maker: function to create a properties dict.

    **Class Methods**

    .. automethod:: get_json_data

    .. automethod:: get_string

    .. versionadded:: 1.3.0
    
    .. _GeoJSON: https://geojson.org/
    """

    def __init__(
        self,
        properties_maker: PropertiesMaker = properties_maker,
        transform_func: TransformFunc = no_transform,
    ) -> None:
        super().__init__()
        self._properties_dict_maker = properties_maker
        self._transform_function = transform_func

    @override
    def get_json_data(self) -> dict[str, Any]:
        """Returns the result as a JSON-like data structure according the GeoJSON specs."""
        if len(self._entities) == 0:
            return {}
        using_features = self._entities[0]["type"] == "Feature"
        if using_features:
            return {"type": "FeatureCollection", "features": self._entities}
        else:
            return {"type": "GeometryCollection", "geometries": self._entities}

    def add_entity(self, entity: dict[str, Any], properties: BackendProperties):
        if not entity:
            return
        properties_dict: dict[str, Any] = self._properties_dict_maker(
            *self.make_properties(properties)
        )
        if properties_dict:
            self._entities.append(
                {
                    "type": "Feature",
                    "properties": properties_dict,
                    "geometry": entity,
                }
            )
        else:
            self._entities.append(entity)

    def make_properties(self, properties: BackendProperties) -> tuple[str, float, str]:
        if self.fixed_lineweight:
            stroke_width = self.fixed_lineweight
        else:
            stroke_width = max(
                self.min_lineweight, properties.lineweight * self.lineweight_scaling
            )
        return (properties.color, round(stroke_width, 2), properties.layer)

    @override
    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.add_entity(
            geojson_object("Point", list(self._transform_function(pos))), properties
        )

    @override
    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        tf = self._transform_function
        self.add_entity(
            geojson_object("LineString", [tf(start), tf(end)]),
            properties,
        )

    @override
    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        lines = list(lines)
        if len(lines) == 0:
            return
        tf = self._transform_function
        json_lines = [(tf(s), tf(e)) for s, e in lines]
        self.add_entity(geojson_object("MultiLineString", json_lines), properties)

    @override
    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        if len(path) == 0:
            return
        tf = self._transform_function
        vertices = [tf(v) for v in path.flattening(distance=self.max_sagitta)]
        self.add_entity(geojson_object("LineString", vertices), properties)

    @override
    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        paths = list(paths)
        if len(paths) == 0:
            return

        polygons: list[GeoJsonPolygon] = []
        for path in paths:
            if len(path):
                polygons.extend(
                    geojson_polygons(
                        path, max_sagitta=self.max_sagitta, tf=self._transform_function
                    )
                )
        if polygons:
            self.add_entity(geojson_object("MultiPolygon", polygons), properties)

    @override
    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        vertices: list[Vec2] = points.vertices()
        if len(vertices) < 3:
            return
        if not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        # exterior ring, without holes
        tf = self._transform_function
        self.add_entity(
            geojson_object("Polygon", [[tf(v) for v in vertices]]), properties
        )


def geojson_object(name: str, coordinates: Any) -> dict[str, Any]:
    return {"type": name, "coordinates": coordinates}


def geojson_ring(
    path: BkPath2d, is_hole: bool, max_sagitta: float, tf: TransformFunc
) -> Ring:
    """Returns a linear ring according to the GeoJSON specs.

    -  A linear ring is a closed LineString with four or more positions.
    -  The first and last positions are equivalent, and they MUST contain
       identical values; their representation SHOULD also be identical.
    -  A linear ring is the boundary of a surface or the boundary of a
       hole in a surface.
    -  A linear ring MUST follow the right-hand rule with respect to the
       area it bounds, i.e., exterior rings are counterclockwise, and
       holes are clockwise.

    """
    if path.has_sub_paths:
        raise TypeError("multi-paths not allowed")
    vertices: Ring = [tf(v) for v in path.flattening(max_sagitta)]
    if not path.is_closed:
        start = path.start
        vertices.append(tf(start))
    clockwise = path.has_clockwise_orientation()
    if (is_hole and not clockwise) or (not is_hole and clockwise):
        vertices.reverse()
    return vertices


def geojson_polygons(
    path: BkPath2d, max_sagitta: float, tf: TransformFunc
) -> list[GeoJsonPolygon]:
    """Returns a list of polygons, where each polygon is a list of an exterior path and
    optional holes e.g. [[ext0, hole0, hole1], [ext1], [ext2, hole0], ...].

    """
    sub_paths: list[BkPath2d] = path.sub_paths()
    if len(sub_paths) == 0:
        return []
    if len(sub_paths) == 1:
        return [[geojson_ring(sub_paths[0], False, max_sagitta, tf)]]

    polygons = nesting.make_polygon_structure(sub_paths)
    geojson_polygons: list[GeoJsonPolygon] = []
    for polygon in polygons:
        geojson_polygon: GeoJsonPolygon = [
            geojson_ring(polygon[0], False, max_sagitta, tf)
        ]  # exterior ring
        if len(polygon) > 1:
            # GeoJSON has no support for nested hole structures, so the sub polygons of
            # holes (hole[1]) are ignored yet!
            holes = polygon[1]
            if isinstance(holes, BkPath2d):  # single hole
                geojson_polygon.append(geojson_ring(holes, True, max_sagitta, tf))
                continue
            if isinstance(holes, (tuple, list)):  # multiple holes
                for hole in holes:
                    if isinstance(hole, (tuple, list)):  # nested polygon
                        # TODO: add sub polygons of holes as separated polygons
                        hole = hole[0]  # exterior path
                    geojson_polygon.append(geojson_ring(hole, True, max_sagitta, tf))

        geojson_polygons.append(geojson_polygon)
    return geojson_polygons



# Copyright (c) 2020-2025, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence, Iterable, Optional, Iterator
from enum import IntEnum
import math
import numpy as np

from ezdxf.math import (
    Vec3,
    Vec2,
    Matrix44,
    X_AXIS,
    Y_AXIS,
    Z_AXIS,
    UVec,
)


__all__ = [
    "is_planar_face",
    "subdivide_face",
    "subdivide_ngons",
    "Plane",
    "PlaneLocationState",
    "normal_vector_3p",
    "safe_normal_vector",
    "distance_point_line_3d",
    "intersection_line_line_3d",
    "intersection_ray_polygon_3d",
    "intersection_line_polygon_3d",
    "basic_transformation",
    "best_fit_normal",
    "BarycentricCoordinates",
    "linear_vertex_spacing",
    "has_matrix_3d_stretching",
    "spherical_envelope",
    "inscribe_circle_tangent_length",
    "bending_angle",
    "split_polygon_by_plane",
    "is_face_normal_pointing_outwards",
    "is_vertex_order_ccw_3d",
]
PI2 = math.pi / 2.0


def is_planar_face(face: Sequence[Vec3], abs_tol=1e-9) -> bool:
    """Returns ``True`` if sequence of vectors is a planar face.

    Args:
         face: sequence of :class:`~ezdxf.math.Vec3` objects
         abs_tol: tolerance for normals check

    """
    if len(face) < 3:
        return False
    if len(face) == 3:
        return True
    first_normal = None
    for index in range(len(face) - 2):
        a, b, c = face[index : index + 3]
        try:
            normal = (b - a).cross(c - b).normalize()
        except ZeroDivisionError:  # colinear edge
            continue
        if first_normal is None:
            first_normal = normal
        elif not first_normal.isclose(normal, abs_tol=abs_tol):
            return False
    if first_normal is not None:
        return True
    return False


def subdivide_face(
    face: Sequence[Vec3], quads: bool = True
) -> Iterator[Sequence[Vec3]]:
    """Subdivides faces by subdividing edges and adding a center vertex.

    Args:
        face: a sequence of :class:`Vec3`
        quads: create quad faces if ``True`` else create triangles

    """
    if len(face) < 3:
        raise ValueError("3 or more vertices required.")

    len_face: int = len(face)
    mid_pos = Vec3.sum(face) / len_face
    subdiv_location: list[Vec3] = [
        face[i].lerp(face[(i + 1) % len_face]) for i in range(len_face)
    ]

    for index, vertex in enumerate(face):
        if quads:
            yield vertex, subdiv_location[index], mid_pos, subdiv_location[index - 1]
        else:
            yield subdiv_location[index - 1], vertex, mid_pos
            yield vertex, subdiv_location[index], mid_pos


def subdivide_ngons(
    faces: Iterable[Sequence[Vec3]],
    max_vertex_count=4,
) -> Iterator[Sequence[Vec3]]:
    """Subdivides faces into triangles by adding a center vertex.

    Args:
        faces: iterable of faces as sequence of :class:`Vec3`
        max_vertex_count: subdivide only ngons with more vertices

    """
    for face in faces:
        if len(face) <= max_vertex_count:
            yield Vec3.tuple(face)
        else:
            mid_pos = Vec3.sum(face) / len(face)
            for index, vertex in enumerate(face):
                yield face[index - 1], vertex, mid_pos


def normal_vector_3p(a: Vec3, b: Vec3, c: Vec3) -> Vec3:
    """Returns normal vector for 3 points, which is the normalized cross
    product for: :code:`a->b x a->c`.
    """
    return (b - a).cross(c - a).normalize()


def safe_normal_vector(vertices: Sequence[Vec3]) -> Vec3:
    """Safe function to detect the normal vector for a face or polygon defined
    by 3 or more `vertices`.

    """
    if len(vertices) < 3:
        raise ValueError("3 or more vertices required")
    a, b, c, *_ = vertices
    try:  # fast path
        return (b - a).cross(c - a).normalize()
    except ZeroDivisionError:  # safe path, can still raise ZeroDivisionError
        return best_fit_normal(vertices)


def best_fit_normal(vertices: Iterable[UVec]) -> Vec3:
    """Returns the "best fit" normal for a plane defined by three or more
    vertices. This function tolerates imperfect plane vertices. Safe function
    to detect the extrusion vector of flat arbitrary polygons.

    """
    # Source: https://gamemath.com/book/geomprims.html#plane_best_fit (9.5.3)
    _vertices = Vec3.list(vertices)
    if len(_vertices) < 3:
        raise ValueError("3 or more vertices required")
    first = _vertices[0]
    if not first.isclose(_vertices[-1]):
        _vertices.append(first)  # close polygon
    prev_x, prev_y, prev_z = first.xyz
    nx = 0.0
    ny = 0.0
    nz = 0.0
    for v in _vertices[1:]:
        x, y, z = v.xyz
        nx += (prev_z + z) * (prev_y - y)
        ny += (prev_x + x) * (prev_z - z)
        nz += (prev_y + y) * (prev_x - x)
        prev_x = x
        prev_y = y
        prev_z = z
    return Vec3(nx, ny, nz).normalize()


def distance_point_line_3d(point: Vec3, start: Vec3, end: Vec3) -> float:
    """Returns the normal distance from a `point` to a 3D line.

    Args:
        point: point to test
        start: start point of the 3D line
        end: end point of the 3D line

    """
    if start.isclose(end):
        raise ZeroDivisionError("Not a line.")
    v1 = point - start
    # point projected onto line start to end:
    v2 = (end - start).project(v1)
    # Pythagoras:
    diff = v1.magnitude_square - v2.magnitude_square
    if diff <= 0.0:
        # This should not happen (abs(v1) > abs(v2)), but floating point
        # imprecision at very small values makes it possible!
        return 0.0
    else:
        return math.sqrt(diff)


def intersection_line_line_3d(
    line1: Sequence[Vec3],
    line2: Sequence[Vec3],
    virtual: bool = True,
    abs_tol: float = 1e-10,
) -> Optional[Vec3]:
    """
    Returns the intersection point of two 3D lines, returns ``None`` if lines
    do not intersect.

    Args:
        line1: first line as tuple of two points as :class:`Vec3` objects
        line2: second line as tuple of two points as :class:`Vec3` objects
        virtual: ``True`` returns any intersection point, ``False`` returns only
            real intersection points
        abs_tol: absolute tolerance for comparisons

    """
    from ezdxf.math import intersection_ray_ray_3d, BoundingBox

    res = intersection_ray_ray_3d(line1, line2, abs_tol)
    if len(res) != 1:
        return None

    point = res[0]
    if virtual:
        return point
    if BoundingBox(line1).inside(point) and BoundingBox(line2).inside(point):
        return point
    return None


def basic_transformation(
    move: UVec = (0, 0, 0),
    scale: UVec = (1, 1, 1),
    z_rotation: float = 0,
) -> Matrix44:
    """Returns a combined transformation matrix for translation, scaling and
    rotation about the z-axis.

    Args:
        move: translation vector
        scale: x-, y- and z-axis scaling as float triplet, e.g. (2, 2, 1)
        z_rotation: rotation angle about the z-axis in radians

    """
    sx, sy, sz = Vec3(scale)
    m = Matrix44.scale(sx, sy, sz)
    if z_rotation:
        m *= Matrix44.z_rotate(z_rotation)
    translate = Vec3(move)
    if not translate.is_null:
        m *= Matrix44.translate(translate.x, translate.y, translate.z)
    return m


PLANE_EPSILON = 1e-9


class PlaneLocationState(IntEnum):
    COPLANAR = 0  # all the vertices are within the plane
    FRONT = 1  # all the vertices are in front of the plane
    BACK = 2  # all the vertices are at the back of the plane
    SPANNING = 3  # some vertices are in front, some in the back


class Plane:
    """Construction tool for 3D planes.

    Represents a plane in 3D space as a normal vector and the perpendicular
    distance from the origin.
    """

    __slots__ = ("_normal", "_distance_from_origin")

    def __init__(self, normal: Vec3, distance: float):
        assert normal.is_null is False, "invalid plane normal"
        self._normal = normal
        # the (perpendicular) distance of the plane from (0, 0, 0)
        self._distance_from_origin = distance

    @property
    def normal(self) -> Vec3:
        """Normal vector of the plane."""
        return self._normal

    @property
    def distance_from_origin(self) -> float:
        """The (perpendicular) distance of the plane from origin (0, 0, 0)."""
        return self._distance_from_origin

    @property
    def vector(self) -> Vec3:
        """Returns the location vector."""
        return self._normal * self._distance_from_origin

    @classmethod
    def from_3p(cls, a: Vec3, b: Vec3, c: Vec3) -> "Plane":
        """Returns a new plane from 3 points in space."""
        try:
            n = (b - a).cross(c - a).normalize()
        except ZeroDivisionError:
            raise ValueError("undefined plane: colinear vertices")
        return Plane(n, n.dot(a))

    @classmethod
    def from_vector(cls, vector: UVec) -> "Plane":
        """Returns a new plane from the given location vector."""
        v = Vec3(vector)
        try:
            return Plane(v.normalize(), v.magnitude)
        except ZeroDivisionError:
            raise ValueError("invalid NULL vector")

    def __copy__(self) -> "Plane":
        """Returns a copy of the plane."""
        return self.__class__(self._normal, self._distance_from_origin)

    copy = __copy__

    def __repr__(self):
        return f"Plane({repr(self._normal)}, {self._distance_from_origin})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Plane):
            return NotImplemented
        return self.vector == other.vector

    def signed_distance_to(self, v: Vec3) -> float:
        """Returns signed distance of vertex `v` to plane, if distance is > 0,
        `v` is in 'front' of plane, in direction of the normal vector, if
        distance is < 0, `v` is at the 'back' of the plane, in the opposite
        direction of the normal vector.

        """
        return self._normal.dot(v) - self._distance_from_origin

    def distance_to(self, v: Vec3) -> float:
        """Returns absolute (unsigned) distance of vertex `v` to plane."""
        return math.fabs(self.signed_distance_to(v))

    def is_coplanar_vertex(self, v: Vec3, abs_tol=1e-9) -> bool:
        """Returns ``True`` if vertex `v` is coplanar, distance from plane to
        vertex `v` is 0.
        """
        return self.distance_to(v) < abs_tol

    def is_coplanar_plane(self, p: "Plane", abs_tol=1e-9) -> bool:
        """Returns ``True`` if plane `p` is coplanar, normal vectors in same or
        opposite direction.
        """
        n_is_close = self._normal.isclose
        return n_is_close(p._normal, abs_tol=abs_tol) or n_is_close(
            -p._normal, abs_tol=abs_tol
        )

    def intersect_line(
        self, start: Vec3, end: Vec3, *, coplanar=True, abs_tol=PLANE_EPSILON
    ) -> Optional[Vec3]:
        """Returns the intersection point of the 3D line from `start` to `end`
        and this plane or ``None`` if there is no intersection. If the argument
        `coplanar` is ``False`` the start- or end point of the line are ignored
        as intersection points.

        """
        state0 = self.vertex_location_state(start, abs_tol)
        state1 = self.vertex_location_state(end, abs_tol)
        if state0 is state1:
            return None
        if not coplanar and (
            state0 is PlaneLocationState.COPLANAR
            or state1 is PlaneLocationState.COPLANAR
        ):
            return None
        n = self.normal
        weight = (self.distance_from_origin - n.dot(start)) / n.dot(end - start)
        return start.lerp(end, weight)

    def intersect_ray(self, origin: Vec3, direction: Vec3) -> Optional[Vec3]:
        """Returns the intersection point of the infinite 3D ray defined by
        `origin` and the `direction` vector and this plane or ``None`` if there
        is no intersection. A coplanar ray does not intersect the plane!

        """
        n = self.normal
        try:
            weight = (self.distance_from_origin - n.dot(origin)) / n.dot(direction)
        except ZeroDivisionError:
            return None
        return origin + (direction * weight)

    def vertex_location_state(
        self, vertex: Vec3, abs_tol=PLANE_EPSILON
    ) -> PlaneLocationState:
        """Returns the :class:`PlaneLocationState` of the given `vertex` in
        relative to this plane.

        """
        distance = self._normal.dot(vertex) - self._distance_from_origin
        if distance < -abs_tol:
            return PlaneLocationState.BACK
        elif distance > abs_tol:
            return PlaneLocationState.FRONT
        else:
            return PlaneLocationState.COPLANAR


def split_polygon_by_plane(
    polygon: Iterable[Vec3],
    plane: Plane,
    *,
    coplanar=True,
    abs_tol=PLANE_EPSILON,
) -> tuple[Sequence[Vec3], Sequence[Vec3]]:
    """Split a convex `polygon` by the given `plane`.

    Returns a tuple of front- and back vertices (front, back).
    Returns also coplanar polygons if the
    argument `coplanar` is ``True``, the coplanar vertices goes into either
    front or back depending on their orientation with respect to this plane.

    """
    polygon_type = PlaneLocationState.COPLANAR
    vertex_types: list[PlaneLocationState] = []
    front_vertices: list[Vec3] = []
    back_vertices: list[Vec3] = []
    vertices = list(polygon)
    w = plane.distance_from_origin
    normal = plane.normal

    # Classify each point as well as the entire polygon into one of four classes:
    # COPLANAR, FRONT, BACK, SPANNING = FRONT + BACK
    for vertex in vertices:
        vertex_type = plane.vertex_location_state(vertex, abs_tol)
        polygon_type |= vertex_type  # type: ignore
        vertex_types.append(vertex_type)

    # Put the polygon in the correct list, splitting it when necessary.
    if polygon_type == PlaneLocationState.COPLANAR:
        if coplanar:
            polygon_normal = best_fit_normal(vertices)
            if normal.dot(polygon_normal) > 0:
                front_vertices = vertices
            else:
                back_vertices = vertices
    elif polygon_type == PlaneLocationState.FRONT:
        front_vertices = vertices
    elif polygon_type == PlaneLocationState.BACK:
        back_vertices = vertices
    elif polygon_type == PlaneLocationState.SPANNING:
        len_vertices = len(vertices)
        for index in range(len_vertices):
            next_index = (index + 1) % len_vertices
            vertex_type = vertex_types[index]
            next_vertex_type = vertex_types[next_index]
            vertex = vertices[index]
            next_vertex = vertices[next_index]
            if vertex_type != PlaneLocationState.BACK:  # FRONT or COPLANAR
                front_vertices.append(vertex)
            if vertex_type != PlaneLocationState.FRONT:  # BACK or COPLANAR
                back_vertices.append(vertex)
            if (vertex_type | next_vertex_type) == PlaneLocationState.SPANNING:
                interpolation_weight = (w - normal.dot(vertex)) / normal.dot(
                    next_vertex - vertex
                )
                plane_intersection_point = vertex.lerp(
                    next_vertex, interpolation_weight
                )
                front_vertices.append(plane_intersection_point)
                back_vertices.append(plane_intersection_point)
        if len(front_vertices) < 3:
            front_vertices = []
        if len(back_vertices) < 3:
            back_vertices = []
    return tuple(front_vertices), tuple(back_vertices)


def intersection_line_polygon_3d(
    start: Vec3,
    end: Vec3,
    polygon: Iterable[Vec3],
    *,
    coplanar=True,
    boundary=True,
    abs_tol=PLANE_EPSILON,
) -> Optional[Vec3]:
    """Returns the intersection point of the 3D line form `start` to `end` and
    the given `polygon`.

    Args:
        start: start point of 3D line as :class:`Vec3`
        end: end point of 3D line as :class:`Vec3`
        polygon: 3D polygon as iterable of :class:`Vec3`
        coplanar: if ``True`` a coplanar start- or end point as intersection
            point is valid
        boundary: if ``True`` an intersection point at the polygon boundary line
            is valid
        abs_tol: absolute tolerance for comparisons

    """
    vertices = list(polygon)
    if len(vertices) < 3:
        raise ValueError("3 or more vertices required")
    try:
        normal = safe_normal_vector(vertices)
    except ZeroDivisionError:
        return None
    plane = Plane(normal, normal.dot(vertices[0]))
    ip = plane.intersect_line(start, end, coplanar=coplanar, abs_tol=abs_tol)
    if ip is None:
        return None
    return _is_intersection_point_inside_3d_polygon(
        ip, vertices, normal, boundary, abs_tol
    )


def intersection_ray_polygon_3d(
    origin: Vec3,
    direction: Vec3,
    polygon: Iterable[Vec3],
    *,
    boundary=True,
    abs_tol=PLANE_EPSILON,
) -> Optional[Vec3]:
    """Returns the intersection point of the infinite 3D ray defined by `origin`
    and the `direction` vector and the given `polygon`.

    Args:
        origin: origin point of the 3D ray as :class:`Vec3`
        direction: direction vector of the 3D ray as :class:`Vec3`
        polygon: 3D polygon as iterable of :class:`Vec3`
        boundary: if ``True`` intersection points at the polygon boundary line
            are valid
        abs_tol: absolute tolerance for comparisons

    """

    vertices = list(polygon)
    if len(vertices) < 3:
        raise ValueError("3 or more vertices required")
    try:
        normal = safe_normal_vector(vertices)
    except ZeroDivisionError:
        return None
    plane = Plane(normal, normal.dot(vertices[0]))
    ip = plane.intersect_ray(origin, direction)
    if ip is None:
        return None
    return _is_intersection_point_inside_3d_polygon(
        ip, vertices, normal, boundary, abs_tol
    )


def _is_intersection_point_inside_3d_polygon(
    ip: Vec3, vertices: list[Vec3], normal: Vec3, boundary: bool, abs_tol: float
):
    from ezdxf.math import is_point_in_polygon_2d, OCS

    ocs = OCS(normal)
    ocs_vertices = Vec2.list(ocs.points_from_wcs(vertices))
    state = is_point_in_polygon_2d(
        Vec2(ocs.from_wcs(ip)), ocs_vertices, abs_tol=abs_tol
    )
    if state > 0 or (boundary and state == 0):
        return ip
    return None


class BarycentricCoordinates:
    """Barycentric coordinate calculation.

    The arguments `a`, `b` and `c` are the cartesian coordinates of an arbitrary
    triangle in 3D space. The barycentric coordinates (b1, b2, b3) define the
    linear combination of `a`, `b` and `c` to represent the point `p`::

        p = a * b1 + b * b2 + c * b3

    This implementation returns the barycentric coordinates of the normal
    projection of `p` onto the plane defined by (a, b, c).

    These barycentric coordinates have some useful properties:

    - if all barycentric coordinates (b1, b2, b3) are in the range [0, 1], then
      the point `p` is inside the triangle (a, b, c)
    - if one of the coordinates is negative, the point `p` is outside the
      triangle
    - the sum of b1, b2 and b3 is always 1
    - the center of "mass" has the barycentric coordinates (1/3, 1/3, 1/3) =
      (a + b + c)/3

    """

    # Source: https://gamemath.com/book/geomprims.html#triangle_barycentric_space

    def __init__(self, a: UVec, b: UVec, c: UVec):
        self.a = Vec3(a)
        self.b = Vec3(b)
        self.c = Vec3(c)
        self._e1 = self.c - self.b
        self._e2 = self.a - self.c
        self._e3 = self.b - self.a
        e1xe2 = self._e1.cross(self._e2)
        self._n = e1xe2.normalize()
        self._denom = e1xe2.dot(self._n)
        if abs(self._denom) < 1e-9:
            raise ValueError("invalid triangle")

    def from_cartesian(self, p: UVec) -> Vec3:
        p = Vec3(p)
        n = self._n
        denom = self._denom
        d1 = p - self.a
        d2 = p - self.b
        d3 = p - self.c
        b1 = self._e1.cross(d3).dot(n) / denom
        b2 = self._e2.cross(d1).dot(n) / denom
        b3 = self._e3.cross(d2).dot(n) / denom
        return Vec3(b1, b2, b3)

    def to_cartesian(self, b: UVec) -> Vec3:
        b1, b2, b3 = Vec3(b).xyz
        return self.a * b1 + self.b * b2 + self.c * b3


def linear_vertex_spacing(start: Vec3, end: Vec3, count: int) -> list[Vec3]:
    """Returns `count` evenly spaced vertices from `start` to `end`."""
    if count <= 2:
        return [start, end]
    distance = end - start
    if distance.is_null:
        return [start] * count

    vertices = [start]
    step = distance.normalize(distance.magnitude / (count - 1))
    for _ in range(1, count - 1):
        start += step
        vertices.append(start)
    vertices.append(end)
    return vertices


def has_matrix_3d_stretching(m: Matrix44) -> bool:
    """Returns ``True`` if matrix `m` performs a non-uniform xyz-scaling.
    Uniform scaling is not stretching in this context.

    Does not check if the target system is a cartesian coordinate system, use the
    :class:`~ezdxf.math.Matrix44` property :attr:`~ezdxf.math.Matrix44.is_cartesian`
    for that.
    """
    ux_mag_sqr = m.transform_direction(X_AXIS).magnitude_square
    uy = m.transform_direction(Y_AXIS)
    uz = m.transform_direction(Z_AXIS)
    return not math.isclose(ux_mag_sqr, uy.magnitude_square) or not math.isclose(
        ux_mag_sqr, uz.magnitude_square
    )


def spherical_envelope(points: Sequence[UVec]) -> tuple[Vec3, float]:
    """Calculate the spherical envelope for the given points.  Returns the
    centroid (a.k.a. geometric center) and the radius of the enclosing sphere.

    .. note::

        The result does not represent the minimal bounding sphere!

    """
    centroid = Vec3.sum(points) / len(points)
    radius = max(centroid.distance(p) for p in points)
    return centroid, radius


def inscribe_circle_tangent_length(dir1: Vec3, dir2: Vec3, radius: float) -> float:
    """Returns the tangent length of an inscribe-circle of the given `radius`.
    The direction `dir1` and `dir2` define two intersection tangents,
    The tangent length is the distance from the intersection point of the
    tangents to the touching point on the inscribe-circle.

    """
    alpha = dir1.angle_between(dir2)
    beta = PI2 - (alpha / 2.0)
    if math.isclose(abs(beta), PI2):
        return 0.0
    return abs(math.tan(beta) * radius)


def bending_angle(dir1: Vec3, dir2: Vec3, normal=Z_AXIS) -> float:
    """Returns the bending angle from `dir1` to `dir2` in radians.

    The normal vector is required to detect the bending orientation,
    an angle > 0 bends to the "left" an angle < 0 bends to the "right".

    """
    angle = dir1.angle_between(dir2)
    nn = dir1.cross(dir2)
    if nn.isclose(normal) or nn.is_null:
        return angle
    elif nn.isclose(-normal):
        return -angle
    raise ValueError("invalid normal vector")


def any_vertex_inside_face(vertices: Sequence[Vec3]) -> Vec3:
    """Returns a vertex from the "inside" of  the given face."""
    # Triangulation is for concave shapes important!
    from ezdxf.math.triangulation import mapbox_earcut_3d

    it = mapbox_earcut_3d(vertices)
    return Vec3.sum(next(it)) / 3.0


def front_faces_intersect_face_normal(
    faces: Sequence[Sequence[Vec3]],
    face: Sequence[Vec3],
    *,
    abs_tol=PLANE_EPSILON,
) -> int:
    """Returns the count of intersections of the normal-vector of the given
    `face` with the `faces` in front of this `face`.

    A counter-clockwise vertex order is assumed!

    """

    def is_face_in_front_of_detector(vertices: Sequence[Vec3]) -> bool:
        if len(vertices) < 3:
            return False
        return any(detector_plane.signed_distance_to(v) > abs_tol for v in vertices)

    # face-normal for counter-clockwise vertex order
    face_normal = safe_normal_vector(face)
    origin = any_vertex_inside_face(face)
    detector_plane = Plane(face_normal, face_normal.dot(origin))

    # collect all faces with at least one vertex in front of the detection plane
    front_faces = (f for f in faces if is_face_in_front_of_detector(f))

    # The detector face is excluded by the
    # is_face_in_front_of_detector() function!

    intersection_points: set[Vec3] = set()
    for face in front_faces:
        ip = intersection_ray_polygon_3d(
            origin, face_normal, face, boundary=True, abs_tol=abs_tol
        )
        if ip is None:
            continue
        if detector_plane.signed_distance_to(ip) > abs_tol:
            # Only count unique intersections points, the ip could lie on an
            # edge (2 ips) or even a corner vertex (3 or more ips).
            intersection_points.add(ip.round(6))
    return len(intersection_points)


def is_face_normal_pointing_outwards(
    faces: Sequence[Sequence[Vec3]],
    face: Sequence[Vec3],
    *,
    abs_tol=PLANE_EPSILON,
) -> bool:
    """Returns ``True`` if the face-normal for the given `face` of a
    closed surface is pointing outwards. A counter-clockwise vertex order is
    assumed, for faces with clockwise vertex order the result is inverted,
    therefore ``False`` is pointing outwards.

    This function does not check if the `faces` are a closed surface.

    """
    return front_faces_intersect_face_normal(faces, face, abs_tol=abs_tol) % 2 == 0


def is_vertex_order_ccw_3d(vertices: list[Vec3], normal: Vec3) -> bool:
    """Returns ``True`` when the given 3D vertices have a counter-clockwise order around
    the given normal vector.

    Works for convex and concave shapes. Does not check or care if all vertices are
    located in a flat plane or if the normal vector is really perpendicular to the
    shape, but the result may be incorrect in that cases.

    Args:
        vertices (list): corner vertices of a flat shape (polygon)
        normal (Vec3): normal vector of the shape

    Raises:
        ValueError: input has less than 3 vertices

    """
    if len(vertices) < 3:
        raise ValueError("3 or more vertices required")

    def signed_area() -> float:
        # using the shoelace formula.
        polygon = np.array(vertices)
        if dom == 0:  # dominant X, use YZ plane
            x, y = polygon[:, 1], polygon[:, 2]
        elif dom == 1:  # dominant Y, use XZ plane
            x, y = polygon[:, 0], polygon[:, 2]
        else:  # dominant Z, use XY plane
            x, y = polygon[:, 0], polygon[:, 1]
        # returns twice the area, but only the sign is relevant for this use case
        return np.dot(x, np.roll(y, -1)) - np.dot(y, np.roll(x, -1))

    # The polygon is maybe concave, direct cross-product checks between
    # adjacent edges are unreliable. Instead, use a projected signed area method.
    # Find dominant axis:
    abs_axis = abs(normal.x), abs(normal.y), abs(normal.z)
    dom = abs_axis.index(max(abs_axis))

    ccw = signed_area() > 0
    dom_axis = normal[dom]
    return dom_axis > 0 if ccw else dom_axis < 0



# Copyright (c) 2019, Manfred Moitzi
# License: MIT License

from io import StringIO
from ezdxf import comments

DXF = """999
preceding comment
0
SECTION
5
ABBA
999
comment before LINE
0
LINE
5
FEFE
999
comment after LINE
0
EOF
"""


def test_load_only_comments():
    stream = StringIO(DXF)
    tags = list(comments.from_stream(stream))
    assert len(tags) == 3
    assert tags[0] == (999, "preceding comment")
    assert tags[1] == (999, "comment before LINE")
    assert tags[2] == (999, "comment after LINE")


def test_load_handles_and_comments():
    stream = StringIO(DXF)
    tags = list(comments.from_stream(stream, codes={5}))
    assert len(tags) == 5
    assert tags[0] == (999, "preceding comment")
    assert tags[1] == (5, "ABBA")
    assert tags[2] == (999, "comment before LINE")
    assert tags[3] == (5, "FEFE")
    # get associated LINE entity:
    # handle = tags[3].value
    # entity = doc.entitydb[handle]
    assert tags[4] == (999, "comment after LINE")


def test_load_structure_and_comments():
    stream = StringIO(DXF)
    tags = list(comments.from_stream(stream, codes={0}))
    assert len(tags) == 6
    assert tags[0] == (999, "preceding comment")
    assert tags[1] == (0, "SECTION")
    assert tags[2] == (999, "comment before LINE")
    assert tags[3] == (0, "LINE")
    assert tags[4] == (999, "comment after LINE")
    assert tags[5] == (0, "EOF")



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence, Union
from typing_extensions import TypeAlias

EncodedData: TypeAlias = Union[str, Sequence[str], bytes, bytearray]



# cython: language_level=3
# Copyright (c) 2021-2024 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Sequence
from .vector cimport Vec3, isclose, v3_dist, v3_lerp, v3_add, Vec2
from .matrix44 cimport Matrix44
import warnings
if TYPE_CHECKING:
    from ezdxf.math import UVec

__all__ = ['Bezier3P']

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL

cdef double RECURSION_LIMIT = 1000


cdef class Bezier3P:
    cdef:
        FastQuadCurve curve  # pyright: ignore
        readonly Vec3 start_point
        Vec3 cp1
        readonly Vec3 end_point

    def __cinit__(self, defpoints: Sequence[UVec]):
        if not isinstance(defpoints[0], (Vec2, Vec3)):
            warnings.warn(
                DeprecationWarning, 
                "Bezier3P requires defpoints of type Vec2 or Vec3 in the future",
            )
        if len(defpoints) == 3:
            self.start_point = Vec3(defpoints[0])
            self.cp1 = Vec3(defpoints[1])
            self.end_point = Vec3(defpoints[2])

            self.curve = FastQuadCurve(
                self.start_point,
                self.cp1,
                self.end_point
            )
        else:
            raise ValueError("Three control points required.")

    @property
    def control_points(self) -> tuple[Vec3, Vec3, Vec3]:
        return self.start_point, self.cp1, self.end_point

    def __reduce__(self):
        return Bezier3P, (self.control_points,)

    def point(self, double t) -> Vec3:
        if 0.0 <= t <= 1.0:
            return self.curve.point(t)
        else:
            raise ValueError("t not in range [0 to 1]")

    def tangent(self, double t) -> Vec3:
        if 0.0 <= t <= 1.0:
            return self.curve.tangent(t)
        else:
            raise ValueError("t not in range [0 to 1]")

    def approximate(self, int segments) -> list[Vec3]:
        cdef double delta_t
        cdef int segment
        cdef list points = [self.start_point]

        if segments < 1:
            raise ValueError(segments)
        delta_t = 1.0 / segments
        for segment in range(1, segments):
            points.append(self.point(delta_t * segment))
        points.append(self.end_point)
        return points

    def flattening(self, double distance, int segments = 4) -> list[Vec3]:
        cdef double dt = 1.0 / segments
        cdef double t0 = 0.0, t1
        cdef _Flattening f = _Flattening(self, distance)
        cdef Vec3 start_point = self.start_point
        cdef Vec3 end_point
        while t0 < 1.0:
            t1 = t0 + dt
            if isclose(t1, 1.0, REL_TOL, ABS_TOL):
                end_point = self.end_point
                t1 = 1.0
            else:
                end_point = self.curve.point(t1)
            f.reset_recursion_check()
            f.flatten(start_point, end_point, t0, t1)
            if f.has_recursion_error():
                raise RecursionError(
                    "Bezier3P flattening error, check for very large coordinates"
                )
            t0 = t1
            start_point = end_point
        return f.points

    def approximated_length(self, segments: int = 128) -> float:
        cdef double length = 0.0
        cdef bint start_flag = 0
        cdef Vec3 prev_point, point

        for point in self.approximate(segments):
            if start_flag:
                length += v3_dist(prev_point, point)
            else:
                start_flag = 1
            prev_point = point
        return length

    def reverse(self) -> Bezier3P:
        return Bezier3P((self.end_point, self.cp1, self.start_point))

    def transform(self, Matrix44 m) -> Bezier3P:
        return Bezier3P(tuple(m.transform_vertices(self.control_points)))


cdef class _Flattening:
    cdef FastQuadCurve curve  # pyright: ignore
    cdef double distance
    cdef list points
    cdef int _recursion_level
    cdef int _recursion_error

    def __cinit__(self, Bezier3P curve, double distance):
        self.curve = curve.curve
        self.distance = distance
        self.points = [curve.start_point]
        self._recursion_level = 0
        self._recursion_error = 0

    cdef has_recursion_error(self):
        return self._recursion_error

    cdef reset_recursion_check(self):
        self._recursion_level = 0
        self._recursion_error = 0

    cdef flatten(
        self,
        Vec3 start_point,
        Vec3 end_point,
        double start_t,
        double end_t
    ):
        if self._recursion_level > RECURSION_LIMIT:
            self._recursion_error = 1
            return
        self._recursion_level += 1
        cdef double mid_t = (start_t + end_t) * 0.5
        cdef Vec3 mid_point = self.curve.point(mid_t)
        cdef double d = v3_dist(mid_point, v3_lerp(start_point, end_point, 0.5))
        if d < self.distance:
            self.points.append(end_point)
        else:
            self.flatten(start_point, mid_point, start_t, mid_t)
            self.flatten(mid_point, end_point, mid_t, end_t)
        self._recursion_level -= 1


cdef class FastQuadCurve:
    cdef:
        double[3] offset
        double[3] p1
        double[3] p2

    def __cinit__(self, Vec3 p0, Vec3 p1, Vec3 p2):
        self.offset[0] = p0.x
        self.offset[1] = p0.y
        self.offset[2] = p0.z

        # 1st control point (p0) is always (0, 0, 0)
        self.p1[0] = p1.x - p0.x
        self.p1[1] = p1.y - p0.y
        self.p1[2] = p1.z - p0.z

        self.p2[0] = p2.x - p0.x
        self.p2[1] = p2.y - p0.y
        self.p2[2] = p2.z - p0.z


    cdef Vec3 point(self, double t):
        # 1st control point (p0) is always (0, 0, 0)
        # => p0 * a is always (0, 0, 0)
        cdef:
            Vec3 result = Vec3()
            # double a = (1 - t) ** 2
            double b = 2.0 * t * (1.0 - t)
            double c = t * t

        iadd_mul(result, self.p1, b)
        iadd_mul(result, self.p2, c)

        # add offset at last - it is maybe very large
        result.x += self.offset[0]
        result.y += self.offset[1]
        result.z += self.offset[2]

        return result


    cdef Vec3 tangent(self, double t):
        # tangent vector is independent from offset location!
        cdef:
            Vec3 result = Vec3()
            # double a = -2 * (1 - t)
            double b = 2.0 - 4.0 * t
            double c = 2.0 * t

        iadd_mul(result, self.p1, b)
        iadd_mul(result, self.p2, c)

        return result


cdef void iadd_mul(Vec3 a, double[3] b, double c):
    a.x += b[0] * c
    a.y += b[1] * c
    a.z += b[2] * c



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.addons.drawing.frontend import _draw_viewports
from ezdxf.entities import Viewport


# The "id" attribute is not evaluated in the current implementation of VIEWPORT
# processing.
def new_viewport(status=0, id=0):
    return Viewport.new(dxfattribs={"status": status, "id": id})


class Frontend:
    def __init__(self):
        self.viewports = []

    def draw_viewport(self, vp):
        self.viewports.append(vp)


@pytest.fixture
def frontend():
    return Frontend()


def test_ignore_active_viewport(frontend):
    # The "active" viewport determines how the paperspace layout is presented as a
    # whole (location & zoom state).
    vp0 = new_viewport(1, 1)  # "active" viewport - ignored
    vp1 = new_viewport(2, 2)  # paperspace viewport - drawn
    _draw_viewports(frontend, [vp0, vp1])
    assert frontend.viewports[0] is vp1


def test_ignore_only_the_first_active_viewport(frontend):
    vp0 = new_viewport(1, 1)  # "active" viewport - ignored
    vp1 = new_viewport(1, 1)  # accidentally second "active" viewport - drawn
    vp2 = new_viewport(2, 2)  # paperspace viewport - drawn
    _draw_viewports(frontend, [vp0, vp1, vp2])
    assert frontend.viewports[0] is vp1
    assert frontend.viewports[1] is vp2


def test_ignore_missing_active_viewport(frontend):
    # No "active" viewport (status=1) defined - this is maybe not a valid paperspace
    # setup!
    vp0 = new_viewport(2, 2)  # paperspace viewport - drawn
    vp1 = new_viewport(3, 3)  # paperspace viewport - drawn
    _draw_viewports(frontend, [vp0, vp1])
    assert frontend.viewports[0] is vp0
    assert frontend.viewports[1] is vp1


def test_ignore_BricsCAD_off_viewports(frontend):
    vp0 = new_viewport(1, 1)  # "active" viewport - ignored
    # BricsCAD set the id to -1 if the viewport is off and 'status' (group code 68)
    # is not present.
    vp1 = Viewport.new(dxfattribs={"id": -1})
    vp2 = new_viewport(3, 3)  # paperspace viewport - drawn
    _draw_viewports(frontend, [vp0, vp1, vp2])
    assert frontend.viewports[0] is vp2


def test_ignore_off_screen_viewports(frontend):
    vp0 = new_viewport(1, 1)  # "active" viewport - ignored
    vp1 = new_viewport(-1, 2)  # off-screen viewport - ignored
    vp2 = new_viewport(2, 3)  # paperspace viewport - drawn
    _draw_viewports(frontend, [vp0, vp1, vp2])
    assert frontend.viewports[0] is vp2


def test_draw_viewports_in_order_of_status(frontend):
    vp0 = new_viewport(1, 1)  # "active" viewport - ignored
    vp1 = new_viewport(status=3, id=2)
    vp2 = new_viewport(status=2, id=3)
    _draw_viewports(frontend, [vp0, vp1, vp2])
    assert frontend.viewports[0] is vp2
    assert frontend.viewports[1] is vp1


if __name__ == "__main__":
    pytest.main([__file__])



# Created: 13.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.lldxf.tags import group_tags, internal_tag_compiler


@pytest.fixture
def groups():
    return list(group_tags(internal_tag_compiler(TESTTAGS)))


def test_init(groups):
    assert 36 == len(groups)


def test_first_group(groups):
    assert "SECTION" == groups[0][0].value
    assert 2 == len(groups[0])


def test_second_group(groups):
    assert "TABLE" == groups[1][0].value
    assert 3 == len(groups[1])


def test_last_group(groups):
    assert "ENDTAB" == groups[-1][0].value
    assert 1 == len(groups[-1])


TESTTAGS = """  0
SECTION
  2
TABLES
  0
TABLE
  2
LTYPE
 70
     5
  0
LTYPE
  2
CONTINUOUS
 70
     0
  3
Solid line
 72
    65
 73
     0
 40
0.0
  0
LTYPE
  2
CENTER
 70
     0
  3
Center ____ _ ____ _ ____ _ ____ _ ____ _ ____
 72
    65
 73
     4
 40
2.0
 49
1.25
 49
-0.25
 49
0.25
 49
-0.25
  0
LTYPE
  2
DASHED
 70
     0
  3
Dashed __ __ __ __ __ __ __ __ __ __ __ __ __ _
 72
    65
 73
     2
 40
0.75
 49
0.5
 49
-0.25
  0
LTYPE
  2
PHANTOM
 70
     0
  3
Phantom ______  __  __  ______  __  __  ______
 72
    65
 73
     6
 40
2.5
 49
1.25
 49
-0.25
 49
0.25
 49
-0.25
 49
0.25
 49
-0.25
  0
LTYPE
  2
HIDDEN
 70
     0
  3
Hidden __ __ __ __ __ __ __ __ __ __ __ __ __ __
 72
    65
 73
     2
 40
9.5249999999999986
 49
6.3499999999999988
 49
-3.1749999999999989
  0
ENDTAB
  0
TABLE
  2
LAYER
 70
     3
  0
LAYER
  2
0
 70
     0
 62
     7
  6
CONTINUOUS
  0
LAYER
  2
VIEW_PORT
 70
     0
 62
     7
  6
CONTINUOUS
  0
LAYER
  2
DEFPOINTS
 70
     0
 62
     7
  6
CONTINUOUS
  0
ENDTAB
  0
TABLE
  2
STYLE
 70
     5
  0
STYLE
  2
STANDARD
 70
     0
 40
0.0
 41
1.0
 50
0.0
 71
     0
 42
0.2
  3
txt
  4

  0
STYLE
  2
ANNOTATIVE
 70
     0
 40
0.0
 41
1.0
 50
0.0
 71
     0
 42
0.2
  3
txt
  4

  0
STYLE
  2
NOTES
 70
     0
 40
3.0
 41
1.0
 50
0.0
 71
     0
 42
0.2
  3
txt
  4

  0
STYLE
  2
TITLES
 70
     0
 40
6.0
 41
1.0
 50
0.0
 71
     0
 42
0.2
  3
txt
  4

  0
STYLE
  2

 70
     1
 40
0.0
 41
1.0
 50
0.0
 71
     0
 42
0.2
  3
ltypeshp.shx
  4

  0
ENDTAB
  0
TABLE
  2
VIEW
 70
     0
  0
ENDTAB
  0
TABLE
  2
UCS
 70
     0
  0
ENDTAB
  0
TABLE
  2
APPID
 70
    10
  0
APPID
  2
ACAD
 70
     0
  0
APPID
  2
ACADANNOPO
 70
     0
  0
APPID
  2
ACADANNOTATIVE
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMJAG
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMTALN
 70
     0
  0
APPID
  2
ACAD_MLEADERVER
 70
     0
  0
APPID
  2
ACAECLAYERSTANDARD
 70
     0
  0
APPID
  2
ACAD_EXEMPT_FROM_CAD_STANDARDS
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMBREAK
 70
     0
  0
APPID
  2
ACAD_PSEXT
 70
     0
  0
ENDTAB
"""



# Copyright (c) 2019-2023 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.image import Image
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

IMAGE = """0
IMAGE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbRasterImage
90
0
10
0.0
20
0.0
30
0.0
11
1.0
21
0.0
31
0.0
12
0.0
22
1.0
32
0.0
13
640
23
320
340
0
70
3
280
0
281
50
282
50
283
0
360
0
71
1
91
2
"""


@pytest.fixture
def entity():
    return Image.from_text(IMAGE)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "IMAGE" in ENTITY_CLASSES


def test_default_init():
    entity = Image()
    assert entity.dxftype() == "IMAGE"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Image.new(
        handle="ABBA", owner="0", dxfattribs={"image_size": (640, 200)}
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.insert is None  # set by add_image()
    assert entity.dxf.u_pixel is None  # set by add_image()
    assert entity.dxf.v_pixel is None  # set by add_image()
    assert entity.dxf.class_version == 0
    assert entity.dxf.image_size == (640, 200)
    assert entity.dxf.image_def_handle is None  # set by add_image()
    assert entity.dxf.flags == 3
    assert entity.dxf.clipping == 0
    assert entity.dxf.brightness == 50
    assert entity.dxf.contrast == 50
    assert entity.dxf.fade == 0
    assert entity.dxf.image_def_reactor_handle is None  # set by add_image()
    assert entity.dxf.clipping_boundary_type == 1
    assert entity.dxf.clip_mode == 0
    assert entity.boundary_path[0] == (-0.5, -0.5)
    assert entity.boundary_path[1] == (639.5, 199.5)


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.dxf.u_pixel == (1, 0, 0)
    assert entity.dxf.v_pixel == (0, 1, 0)
    assert entity.dxf.class_version == 0
    assert entity.dxf.image_size == (640, 320)
    assert entity.dxf.image_def_handle == "0"
    assert entity.dxf.flags == 3
    assert entity.dxf.clipping == 0
    assert entity.dxf.brightness == 50
    assert entity.dxf.contrast == 50
    assert entity.dxf.fade == 0
    assert entity.dxf.image_def_reactor_handle == "0"
    assert entity.dxf.clipping_boundary_type == 1
    assert entity.dxf.clip_mode == 0
    assert entity.boundary_path[0] == (-0.5, -0.5)
    assert entity.boundary_path[1] == (639.5, 319.5)


def test_image_write_dxf():
    entity = Image.from_text(IMAGE)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(EXPECTED)
    assert result == expected


EXPECTED = """0
IMAGE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbRasterImage
90
0
10
0.0
20
0.0
30
0.0
11
1.0
21
0.0
31
0.0
12
0.0
22
1.0
32
0.0
13
640
23
320
340
0
70
3
280
0
281
50
282
50
283
0
360
0
71
1
91
2
14
-0.5
24
-0.5
14
639.5
24
319.5
290
0
"""



import pathlib
import sys
from abc import ABC, abstractmethod
import subprocess
import os
import platform

from ezdxf.addons.drawing.backend import BackendInterface


class FileOutputRenderBackend(ABC):
    def __init__(self, dpi: float) -> None:
        self._dpi = dpi

    @abstractmethod
    def supported_formats(self) -> list[tuple[str, str]]:
        raise NotImplementedError

    @abstractmethod
    def default_format(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def backend(self) -> BackendInterface:
        raise NotImplementedError

    @abstractmethod
    def save(self, output: pathlib.Path) -> None:
        raise NotImplementedError


class MatplotlibFileOutput(FileOutputRenderBackend):
    def __init__(self, dpi: float) -> None:
        super().__init__(dpi)

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("Matplotlib not found") from None

        from ezdxf.addons.drawing.matplotlib import MatplotlibBackend

        self._plt = plt
        self._fig = plt.figure()
        self._ax = self._fig.add_axes((0, 0, 1, 1))
        self._backend = MatplotlibBackend(self._ax)

    def supported_formats(self) -> list[tuple[str, str]]:
        return list(self._fig.canvas.get_supported_filetypes().items())

    def default_format(self) -> str:
        return "png"

    def backend(self) -> BackendInterface:
        return self._backend

    def save(self, output: pathlib.Path) -> None:
        self._fig.savefig(output, dpi=self._dpi)
        self._plt.close(self._fig)


class PyQtFileOutput(FileOutputRenderBackend):
    def __init__(self, dpi: float) -> None:
        super().__init__(dpi)

        try:
            from ezdxf.addons.xqt import QtCore, QtGui, QtWidgets
            from ezdxf.addons.drawing.pyqt import PyQtBackend
        except ImportError:
            raise ImportError("PyQt not found") from None

        self._qc = QtCore
        self._qg = QtGui
        self._qw = QtWidgets
        self._app = QtWidgets.QApplication(sys.argv)
        self._scene = QtWidgets.QGraphicsScene()
        self._backend = PyQtBackend()
        self._backend.set_scene(self._scene)

    def supported_formats(self) -> list[tuple[str, str]]:
        # https://doc.qt.io/qt-6/qimage.html#reading-and-writing-image-files
        return [
            ("bmp", "Windows Bitmap"),
            ("jpg", "Joint Photographic Experts Group"),
            ("jpeg", "Joint Photographic Experts Group"),
            ("png", "Portable Network Graphics"),
            ("ppm", "Portable Pixmap"),
            ("xbm", "X11 Bitmap"),
            ("xpm", "X11 Pixmap"),
            ("svg", "Scalable Vector Graphics"),
        ]

    def default_format(self) -> str:
        return "png"

    def backend(self) -> BackendInterface:
        return self._backend

    def save(self, output: pathlib.Path) -> None:
        if output.suffix.lower() == ".svg":
            from PySide6.QtSvg import QSvgGenerator

            generator = QSvgGenerator()
            generator.setFileName(str(output))
            generator.setResolution(int(self._dpi))
            scene_rect = self._scene.sceneRect()
            output_size = self._qc.QSize(
                round(scene_rect.size().width()), round(scene_rect.size().height())
            )
            generator.setSize(output_size)
            generator.setViewBox(
                self._qc.QRect(0, 0, output_size.width(), output_size.height())
            )

            painter = self._qg.QPainter()

            transform = self._qg.QTransform()
            transform.scale(1, -1)
            transform.translate(0, -output_size.height())

            painter.begin(generator)
            painter.setWorldTransform(transform, combine=True)
            painter.setRenderHint(self._qg.QPainter.RenderHint.Antialiasing)
            self._scene.render(painter)
            painter.end()

        else:

            view = self._qw.QGraphicsView(self._scene)
            view.setRenderHint(self._qg.QPainter.RenderHint.Antialiasing)
            sizef: QRectF = self._scene.sceneRect() * self._dpi / 92  # type: ignore
            image = self._qg.QImage(
                self._qc.QSize(round(sizef.width()), round(sizef.height())),
                self._qg.QImage.Format.Format_ARGB32,
            )
            painter = self._qg.QPainter(image)
            painter.setRenderHint(self._qg.QPainter.RenderHint.Antialiasing)
            painter.fillRect(image.rect(), self._scene.backgroundBrush())
            self._scene.render(painter)
            painter.end()
            image.mirror(False, True)
            image.save(str(output))


class MuPDFFileOutput(FileOutputRenderBackend):
    def __init__(self, dpi: float) -> None:
        super().__init__(dpi)

        from ezdxf.addons.drawing.pymupdf import PyMuPdfBackend, is_pymupdf_installed

        if not is_pymupdf_installed:
            raise ImportError("PyMuPDF not found")
        self._backend = PyMuPdfBackend()

    def supported_formats(self) -> list[tuple[str, str]]:
        # https://pymupdf.readthedocs.io/en/latest/pixmap.html#pixmapoutput
        return [
            ("pdf", "Portable Document Format"),
            ("svg", "Scalable Vector Graphics"),
            ("jpg", "Joint Photographic Experts Group"),
            ("jpeg", "Joint Photographic Experts Group"),
            ("pam", "Portable Arbitrary Map"),
            ("pbm", "Portable Bitmap"),
            ("pgm", "Portable Graymap"),
            ("png", "Portable Network Graphics"),
            ("pnm", "Portable Anymap"),
            ("ppm", "Portable Pixmap (no alpha channel)"),
            ("ps", "Adobe PostScript Image"),
            ("psd", "Adobe Photoshop Document"),
        ]

    def default_format(self) -> str:
        return "pdf"

    def backend(self) -> BackendInterface:
        return self._backend

    def save(self, output: pathlib.Path) -> None:
        from ezdxf.addons.drawing import layout

        backend = self._backend.get_replay(layout.Page(0, 0))
        if output.suffix == ".pdf":
            output.write_bytes(backend.get_pdf_bytes())
        elif output.suffix == ".svg":
            output.write_text(backend.get_svg_image())
        else:
            pixmap = backend.get_pixmap(int(self._dpi), alpha=True)
            pixmap.save(str(output))


class SvgFileOutput(FileOutputRenderBackend):
    def __init__(self, dpi: float) -> None:
        super().__init__(dpi)

        from ezdxf.addons.drawing.svg import SVGBackend

        self._backend = SVGBackend()

    def supported_formats(self) -> list[tuple[str, str]]:
        return [("svg", "Scalable Vector Graphics")]

    def default_format(self) -> str:
        return "svg"

    def backend(self) -> BackendInterface:
        return self._backend

    def save(self, output: pathlib.Path) -> None:
        from ezdxf.addons.drawing import layout

        output.write_text(self._backend.get_string(layout.Page(0, 0)))


def open_file(path: pathlib.Path) -> None:
    """open the given path in the default application"""
    system = platform.system()
    if system == "Darwin":
        subprocess.call(
            ["open", str(path)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
    elif system == "Windows":
        os.startfile(str(path))  # type: ignore
    else:
        subprocess.call(
            ["xdg-open", str(path)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.render.trace import TraceBuilder, LinearTrace, CurvedTrace
from ezdxf.math import BSpline, Vec2, BoundingBox


def test_trace_builder_init():
    t = TraceBuilder()
    assert len(t) == 0


def test_add_station_2d():
    t = LinearTrace()
    t.add_station((0, 0), 1, 2)
    assert len(t) == 1
    assert t[0].vertex == (0, 0)
    assert t[0].start_width == 1
    assert t[0].end_width == 2
    t.add_station((1, 0), 1)
    assert len(t) == 2
    assert t[1].vertex == (1, 0)
    assert t[1].start_width == 1
    assert t[1].end_width == 1


def test_add_station_3d():
    # z-axis is ignored
    t = LinearTrace()
    t.add_station((0, 0, 0), 1, 2)
    assert t[0].vertex == (0, 0)


def test_add_spline_segment():
    t = CurvedTrace.from_spline(
        BSpline.from_fit_points([(1, 0), (3, 1), (5, -1), (6, 0)]),
        start_width=2,
        end_width=1,
        segments=10,
    )
    assert len(t) == 11


def test_square_face():
    t = LinearTrace()
    t.add_station((0, 0), 1, 1)
    t.add_station((1, 0), 0, 0)
    face = list(t.faces())[0]
    assert face[0].isclose(Vec2(0, +0.5))
    assert face[1].isclose(Vec2(0, -0.5))
    assert face[2].isclose(Vec2(1, -0.5))
    assert face[3].isclose(Vec2(1, +0.5))


def test_closed_linear_path():
    t = LinearTrace()
    t.add_station((0, 0), 1, 1)
    t.add_station((1, 0), 1, 1)
    t.add_station((1, 1), 1, 1)
    t.add_station((0, 1), 1, 1)
    t.add_station((0, 0), 1, 1)
    faces = list(t.faces())
    assert len(faces) == 4
    assert faces[0] == (
        Vec2(0.5, 0.5),
        Vec2(-0.5, -0.5),
        Vec2(1.5, -0.5),
        Vec2(0.5, 0.5),
    )
    assert faces[3] == (
        Vec2(0.5, 0.5),
        Vec2(-0.5, 1.5),
        Vec2(-0.5, -0.5),
        Vec2(0.5, 0.5),
    )


def test_two_straight_faces():
    t = LinearTrace()
    t.add_station((0, 0), 1, 1)
    t.add_station((2, 0), 1, 1)
    t.add_station((4, 0), 1, 1)
    face1, face2 = list(t.faces())
    assert face1[0].isclose(Vec2(0, +0.5))
    assert face1[1].isclose(Vec2(0, -0.5))
    assert face1[2].isclose(Vec2(2, -0.5))
    assert face1[3].isclose(Vec2(2, +0.5))
    assert face2[0].isclose(Vec2(2, +0.5))
    assert face2[1].isclose(Vec2(2, -0.5))
    assert face2[2].isclose(Vec2(4, -0.5))
    assert face2[3].isclose(Vec2(4, +0.5))


def test_two_angled_faces():
    t = LinearTrace()
    t.add_station((0, 0), 1, 0.5)
    t.add_station((2, 0), 1, 1)
    t.add_station((4, 2), 1, 1)
    face1, face2 = list(t.faces())
    assert face1[0].isclose(Vec2(0, +0.5))
    assert face1[1].isclose(Vec2(0, -0.5))
    assert face1[2].isclose(Vec2(2.5224077499274835, -0.18469903125906456))
    assert face1[3].isclose(Vec2(1.5936828611675133, 0.3007896423540608))
    assert face2[2].isclose(Vec2(4.353553390593274, 1.6464466094067263))
    assert face2[3].isclose(Vec2(3.646446609406726, 2.353553390593274))


def test_linear_trace_polygon():
    t = LinearTrace()
    t.add_station((0, 0), 1, 1)
    t.add_station((2, 0), 1, 1)
    t.add_station((4, 0), 1, 1)
    polygon = t.polygon()
    assert len(polygon) == 6
    assert polygon[0].isclose(Vec2(0, -0.5))
    assert polygon[-1].isclose(Vec2(0, +0.5))


def test_virtual_entities_added_to_entity_database():
    doc = ezdxf.new()
    msp = doc.modelspace()
    t = LinearTrace()
    t.add_station((0, 0), 1, 1)
    t.add_station((1, 0), 0, 0)

    dxfattribs = {"layer": "TEST"}
    solid = list(t.virtual_entities("SOLID", dxfattribs, doc))[0]
    assert solid.DXFTYPE == "SOLID"
    assert solid.dxf.layer == "TEST"
    assert solid.dxf.handle in doc.entitydb
    msp.add_entity(solid)

    trace = list(t.virtual_entities("TRACE", dxfattribs, doc))[0]
    assert trace.DXFTYPE == "TRACE"
    assert trace.dxf.layer == "TEST"
    assert trace.dxf.handle in doc.entitydb
    msp.add_entity(trace)

    face = list(t.virtual_entities("3DFACE", dxfattribs, doc))[0]
    assert face.DXFTYPE == "3DFACE"
    assert face.dxf.layer == "TEST"
    assert face.dxf.handle in doc.entitydb
    msp.add_entity(face)

    assert len(msp) == 3


def test_issue_191():
    from ezdxf.entities import factory

    e = factory.new(
        "LWPOLYLINE",
        dxfattribs={
            "flags": 1,
        },
    )
    e.set_points(
        [
            (421846.9857097387, -36908.41493252141, 0.0, 50.0, 0.52056705),
            (421846.9857097387, -36908.41493252139, 0.0, 50.0, 0.52056705),
        ]
    )
    trace = TraceBuilder.from_polyline(e, segments=64)
    assert len(trace) == 1
    assert len(list(trace.faces())) == 32


def test_nearly_horizontal_parallel_lines_in_linear_trace_builder():
    width = 30
    trace = LinearTrace()
    trace.add_station((100, 0), width, width)
    trace.add_station((100_000, 0), width, width)
    trace.add_station((0, 0.001), width, width)
    bbox = BoundingBox()
    for face in trace.faces():
        bbox.extend(face)
    assert bbox.extmin.x > -100


def test_nearly_vertical_parallel_lines_in_linear_trace_builder():
    width = 30
    trace = LinearTrace()
    trace.add_station((0, 100), width, width)
    trace.add_station((0, 100_000), width, width)
    trace.add_station((0.001, 0), width, width)
    bbox = BoundingBox()
    for face in trace.faces():
        bbox.extend(face)
    assert bbox.extmin.y > -100


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright 2018, Manfred Moitzi
# License: MIT License
from __future__ import unicode_literals
import pytest
import os
import random
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new
from ezdxf.math import Vec3


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param, setup=True)


def test_linear_dimline_all_arrow_style(drawing, tmpdir):
    dwg = drawing
    msp = dwg.modelspace()
    ezdxf_dimstyle = dwg.dimstyles.get("EZDXF")
    ezdxf_dimstyle.copy_to_header(dwg)

    for index, name in enumerate(sorted(ezdxf.ARROWS.__all_arrows__)):
        y = index * 4
        attributes = {
            "dimtxsty": "LiberationMono",
            "dimblk": name,
            "dimtsz": 0.0,
            "dimdle": 0.5,
            "dimasz": 0.25,
        }

        dim = msp.add_linear_dim(
            base=(3, y + 2),
            p1=(0, y),
            p2=(3, y),
            dimstyle="EZDXF",
            override=attributes,
        )
        dim.render()

    filename = str(
        tmpdir.join("all_dimline_styles_{}.dxf".format(drawing.dxfversion))
    )
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)


def random_point(start, end):
    dist = end - start
    return Vec3(start + random.random() * dist, start + random.random() * dist)


def test_random_multi_point_linear_dimension(tmpdir):
    length = 20
    count = 10
    fname = "multi_random_point_linear_dim_R2007.dxf"

    dwg = ezdxf.new("R2007", setup=True)
    msp = dwg.modelspace()
    points = [random_point(0, length) for _ in range(count)]
    msp.add_lwpolyline(points, dxfattribs={"color": 1})

    # create quick a new DIMSTYLE as alternative to overriding DIMSTYLE attributes
    dimstyle = dwg.dimstyles.duplicate_entry("EZDXF", "WITHTFILL")

    dimstyle.dxf.dimtfill = 1
    dimstyle.dxf.dimdec = 2

    dimstyle = dwg.dimstyles.duplicate_entry("WITHTFILL", "WITHTXT")
    dimstyle.dxf.dimblk = ezdxf.ARROWS.closed
    dimstyle.dxf.dimtxsty = "STANDARD"

    msp.add_multi_point_linear_dim(
        base=(0, length + 2), points=points, dimstyle="WITHTFILL"
    )
    msp.add_multi_point_linear_dim(
        base=(-2, 0), points=points, angle=90, dimstyle="WITHTFILL"
    )
    msp.add_multi_point_linear_dim(
        base=(10, -10), points=points, angle=45, dimstyle="WITHTXT"
    )

    filename = str(tmpdir.join(fname))
    try:
        dwg.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail("DXFError: {0} for {1}".format(str(e), fname))
    assert os.path.exists(filename)


def test_draw_all_arrows(drawing, tmpdir):
    msp = drawing.modelspace()
    y = 0
    for index, name in enumerate(sorted(ezdxf.ARROWS.__all_arrows__)):
        if name == "":
            label = '"" = closed filled'
        else:
            label = name
        y = index * 2

        def add_connection_point(p):
            msp.add_circle(p, radius=0.01, dxfattribs={"color": 1})

        msp.add_text(
            label, dxfattribs={"style": "OpenSans", "height": 0.25}
        ).set_placement((-5, y - 0.5))
        msp.add_line((-5, y), (-1, y))
        msp.add_line((5, y), (10, y))
        # left side |<- is the reverse orientation
        cp1 = msp.add_arrow(name, insert=(0, y), size=1, rotation=180)
        # right side ->| is the base orientation
        cp2 = msp.add_arrow(name, insert=(4, y), size=1, rotation=0)
        msp.add_line(cp1, cp2)
        add_connection_point(cp1)
        add_connection_point(cp2)

        add_connection_point(
            msp.add_arrow_blockref(name, insert=(7, y), size=0.3, rotation=45)
        )
        add_connection_point(
            msp.add_arrow_blockref(
                name, insert=(7.5, y), size=0.3, rotation=135
            )
        )
        add_connection_point(
            msp.add_arrow_blockref(name, insert=(8, y), size=0.5, rotation=-90)
        )

    msp.add_line((0, 0), (0, y))
    msp.add_line((4, 0), (4, y))
    msp.add_line((8, 0), (8, y))

    filename = str(
        tmpdir.join("draw_all_arrows_{}.dxf".format(drawing.dxfversion))
    )
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)
